from __future__ import annotations

import threading
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Generator

    from tests.config.loader.conftest import ReplaceOne
    from tox.pytest import LogCaptureFixture, MonkeyPatch


def test_replace_env_set(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    monkeypatch.setenv("MAGIC", "something good")
    result = replace_one("{env:MAGIC}")
    assert result == "something good"


def test_replace_env_set_double_bs(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Double backslash should remain but not affect surrounding replacements."""
    monkeypatch.setenv("MAGIC", "something good")
    result = replace_one(r"{env:MAGIC}\\{env:MAGIC}")
    assert result == r"something good\\something good"


def test_replace_env_set_triple_bs(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Triple backslash should retain two slashes with the third escaping subsequent replacement."""
    monkeypatch.setenv("MAGIC", "something good")
    result = replace_one(r"{env:MAGIC}\\\{env:MAGIC}")
    assert result == r"something good\\{env:MAGIC}"


def test_replace_env_set_quad_bs(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Quad backslash should remain but not affect surrounding replacements."""
    monkeypatch.setenv("MAGIC", "something good")
    result = replace_one(r"\\{env:MAGIC}\\\\{env:MAGIC}" + "\\")
    assert result == r"\\something good\\\\something good" + "\\"


def test_replace_env_when_value_is_backslash(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """When the replacement value is backslash, it shouldn't affect the next replacement."""
    monkeypatch.setenv("MAGIC", "tragic")
    monkeypatch.setenv("BS", "\\")
    result = replace_one(r"{env:BS}{env:MAGIC}")
    assert result == r"\tragic"


def test_replace_env_when_value_is_stuff_then_backslash(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """When the replacement value is a string containing backslash, it shouldn't affect the next replacement."""
    monkeypatch.setenv("MAGIC", "tragic")
    monkeypatch.setenv("BS", "stuff\\")
    result = replace_one(r"{env:BS}{env:MAGIC}")
    assert result == r"stuff\tragic"


def test_replace_env_missing(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    monkeypatch.delenv("MAGIC", raising=False)
    result = replace_one("{env:MAGIC}")
    assert not result


def test_replace_env_missing_default(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    monkeypatch.delenv("MAGIC", raising=False)
    result = replace_one("{env:MAGIC:def}")
    assert result == "def"


def test_replace_env_missing_default_from_env(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    monkeypatch.delenv("MAGIC", raising=False)
    monkeypatch.setenv("MAGIC_DEFAULT", "yes")
    result = replace_one("{env:MAGIC:{env:MAGIC_DEFAULT}}")
    assert result == "yes"


def test_replace_env_var_multiple(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Multiple env substitutions on a single line."""
    monkeypatch.setenv("MAGIC", "MAGIC")
    monkeypatch.setenv("TRAGIC", "TRAGIC")
    result = replace_one("{env:MAGIC} {env:TRAGIC} {env:MAGIC}")
    assert result == "MAGIC TRAGIC MAGIC"


def test_replace_env_var_multiple_default(replace_one: ReplaceOne) -> None:
    """Multiple env substitutions on a single line with default values."""
    result = replace_one("{env:MAGIC:foo} {env:TRAGIC:bar} {env:MAGIC:baz}")
    assert result == "foo bar baz"


def test_replace_env_var_circular(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Replacement values will not infinitely loop"""
    monkeypatch.setenv("MAGIC", "{env:MAGIC}")
    result = replace_one("{env:MAGIC}")
    assert result == "{env:MAGIC}"


@pytest.fixture
def reset_env_var_after_delay(monkeypatch: MonkeyPatch) -> Generator[threading.Thread, None, None]:
    timeout = 2

    def avoid_infinite_loop() -> None:  # pragma: no cover
        monkeypatch.setenv("TRAGIC", f"envvar forcibly reset after {timeout} sec")

    timer = threading.Timer(2, avoid_infinite_loop)
    timer.start()
    yield timer
    timer.cancel()
    timer.join()


@pytest.mark.usefixtures("reset_env_var_after_delay")
def test_replace_env_var_circular_flip_flop(
    replace_one: ReplaceOne,
    monkeypatch: MonkeyPatch,
    caplog: LogCaptureFixture,
) -> None:
    """Replacement values will not infinitely loop back and forth"""
    monkeypatch.setenv("TRAGIC", "{env:MAGIC}")
    monkeypatch.setenv("MAGIC", "{env:TRAGIC}")
    result = replace_one("{env:MAGIC}")
    assert result == "{env:MAGIC}"
    assert "circular chain between set env MAGIC, TRAGIC" in caplog.messages


@pytest.mark.usefixtures("reset_env_var_after_delay")
def test_replace_env_var_circular_flip_flop_5(
    replace_one: ReplaceOne,
    monkeypatch: MonkeyPatch,
    caplog: LogCaptureFixture,
) -> None:
    """Replacement values will not infinitely loop back and forth (longer chain)"""
    monkeypatch.setenv("MAGIC", "{env:TRAGIC}")
    monkeypatch.setenv("TRAGIC", "{env:RABBIT}")
    monkeypatch.setenv("RABBIT", "{env:HAT}")
    monkeypatch.setenv("HAT", "{env:TRICK}")
    monkeypatch.setenv("TRICK", "{env:MAGIC}")
    result = replace_one("{env:MAGIC}")
    assert result == "{env:MAGIC}"
    assert "circular chain between set env MAGIC, TRAGIC, RABBIT, HAT, TRICK" in caplog.messages


@pytest.mark.parametrize("fallback", [True, False])
def test_replace_env_var_chase(replace_one: ReplaceOne, monkeypatch: MonkeyPatch, fallback: bool) -> None:
    """Resolve variable to be replaced and default value via indirection."""
    monkeypatch.setenv("WALK", "THIS")
    def_val = "or that one"
    monkeypatch.setenv("DEF", def_val)
    if fallback:
        monkeypatch.delenv("THIS", raising=False)
        exp_result = def_val
    else:
        this_val = "path"
        monkeypatch.setenv("THIS", this_val)
        exp_result = this_val
    result = replace_one("{env:{env:WALK}:{env:DEF}}")
    assert result == exp_result


def test_replace_env_default_with_colon(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    monkeypatch.delenv("MAGIC", raising=False)
    result = replace_one("{env:MAGIC:https://some.url.org}")
    assert result == "https://some.url.org"


def test_replace_env_default_deep(replace_one: ReplaceOne, monkeypatch: MonkeyPatch) -> None:
    """Get the value through a long tree of nested defaults."""
    monkeypatch.delenv("M", raising=False)
    assert replace_one("{env:M:{env:M:{env:M:{env:M:{env:M:foo}}}}}") == "foo"



from __future__ import annotations

import os
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from tox.tox_env.api import ToxEnv

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


def test_recreate(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\nrecreate=True"})
    result_first = prj.run("r")
    result_first.assert_success()

    result_second = prj.run("r")
    result_second.assert_success()
    assert "remove tox env folder" in result_second.out


def test_allow_list_external_fail(tox_project: ToxProjectCreator, fake_exe_on_path: Path) -> None:
    prj = tox_project({"tox.ini": f"[testenv]\npackage=skip\ncommands={fake_exe_on_path.stem}"})
    execute_calls = prj.patch_execute(lambda r: 0 if "cmd" in r.run_id else None)

    result = prj.run("r")

    result.assert_failed(1)
    out = rf".*py: failed with {fake_exe_on_path.stem} is not allowed, use allowlist_externals to allow it.*"
    result.assert_out_err(out=out, err="", regex=True)
    execute_calls.assert_called()


def test_env_log(tox_project: ToxProjectCreator) -> None:
    cmd = "commands=python -c 'import sys; print(1); print(2); print(3, file=sys.stderr); print(4, file=sys.stderr)'"
    env_vars = "    UNPREDICTABLE = 🪟"
    prj = tox_project({"tox.ini": f"[testenv]\npackage=skip\nset_env =\n{env_vars}\n{cmd}"})
    result_first = prj.run("r")
    result_first.assert_success()

    log_dir = prj.path / ".tox" / "py" / "log"
    assert log_dir.exists(), result_first.out

    filename = {i.name for i in log_dir.iterdir()}
    assert filename == {"1-commands[0].log"}
    content = (log_dir / "1-commands[0].log").read_text()

    assert f"cwd: {prj.path}" in content
    assert f"allow: {prj.path}" in content
    assert "metadata " in content
    assert "env PATH: " in content
    assert content.startswith("name: py\nrun_id: commands[0]")
    assert "cmd: python -c" in content
    ending = """
    exit_code: 0
    1
    2

    standard error:
    3
    4
    """
    assert content.endswith(dedent(ending).lstrip()), content

    result_second = prj.run("r")  # second run overwrites, so no new files
    result_second.assert_success()
    filename = {i.name for i in log_dir.iterdir()}
    assert filename == {"1-commands[0].log"}


def test_tox_env_pass_env_literal_exist() -> None:
    with patch("os.environ", {"A": "1"}):
        env = ToxEnv._load_pass_env(["A"])  # noqa: SLF001
    assert env == {"A": "1"}


def test_tox_env_pass_env_literal_miss() -> None:
    with patch("os.environ", {}):
        env = ToxEnv._load_pass_env(["A"])  # noqa: SLF001
    assert not env


def test_tox_env_pass_env_fails_on_whitespace(tox_project: ToxProjectCreator) -> None:
    first, second = "A B", "C D"
    prj = tox_project({"tox.ini": f"[testenv]\npackage=skip\npass_env = {first}\n {second}\n  E"})
    result = prj.run("c", "-k", "pass_env", raise_on_config_fail=False)
    result.assert_failed(code=-1)
    msg = (
        '[testenv:py]\npass_env = # Exception: Fail("pass_env values cannot contain whitespace, use comma to have '
        f'multiple values in a single line, invalid values found {first!r}, {second!r}")\n'
    )
    assert result.out == msg

    result = prj.run("r")
    result.assert_failed(1)
    msg = (
        "py: failed with pass_env values cannot contain whitespace, use comma to have multiple values in a single line,"
        " invalid values found 'A B', 'C D'"
    )
    assert msg in result.out


@pytest.mark.parametrize("glob", ["*", "?"])
@pytest.mark.parametrize("char", ["a", "A"])
def test_tox_env_pass_env_match_ignore_case(char: str, glob: str) -> None:
    with patch("os.environ", {"A1": "1", "a2": "2", "A2": "3", "B": "4"}):
        env = ToxEnv._load_pass_env([f"{char}{glob}"])  # noqa: SLF001
    assert env == {"A1": "1", "a2": "2", "A2": "3"}


def test_disallow_pass_env(tox_project: ToxProjectCreator) -> None:
    cmd = "import os; print(os.environ.get('FOO_BAR', '')); print(os.environ.get('FOO_SECRET', ''))"
    toml = f"""
    [env_run_base]
    package = "skip"
    pass_env = ["FOO_*"]
    disallow_pass_env = ["FOO_SECRET"]
    commands = [["python", "-c", "{cmd}"]]
    """
    project = tox_project({"tox.toml": toml})
    with patch.dict("os.environ", {"FOO_BAR": "visible", "FOO_SECRET": "hidden"}):
        result = project.run("r")
    result.assert_success()
    assert "visible" in result.out
    assert "hidden" not in result.out


def test_disallow_pass_env_glob(tox_project: ToxProjectCreator) -> None:
    cmd = "import os; print(os.environ.get('APP_NAME', '')); print(os.environ.get('APP_SECRET_KEY', ''))"
    toml = f"""
    [env_run_base]
    package = "skip"
    pass_env = ["APP_*"]
    disallow_pass_env = ["APP_SECRET_*"]
    commands = [["python", "-c", "{cmd}"]]
    """
    project = tox_project({"tox.toml": toml})
    with patch.dict("os.environ", {"APP_NAME": "myapp", "APP_SECRET_KEY": "s3cret"}):
        result = project.run("r")
    result.assert_success()
    assert "myapp" in result.out
    assert "s3cret" not in result.out


def test_disallow_pass_env_empty(tox_project: ToxProjectCreator) -> None:
    toml = """
    [env_run_base]
    package = "skip"
    pass_env = ["FOO"]
    commands = [["python", "-c", "import os; print(os.environ.get('FOO', ''))"]]
    """
    project = tox_project({"tox.toml": toml})
    with patch.dict("os.environ", {"FOO": "bar"}):
        result = project.run("r")
    result.assert_success()
    assert "bar" in result.out


def test_change_dir_posargs_no_recursion(tox_project: ToxProjectCreator) -> None:
    toml = """\
[env.foo]
package = "skip"
change_dir = "{posargs}"
commands = [["python", "--version"]]
"""
    prj = tox_project({"tox.toml": toml})
    (prj.path / "subdir").mkdir()
    result = prj.run("r", "-e", "foo", "--", "subdir")
    result.assert_success()
    assert f"foo: commands[0] {prj.path / 'subdir'}>" in result.out


def test_posargs_cross_drive_no_crash(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    toml = """\
[env_run_base]
package = "skip"
commands = [["python", "{posargs}"]]
"""
    prj = tox_project({"tox.toml": toml})
    (prj.path / "test.py").write_text("print('ok')")
    original_relpath = os.path.relpath

    def _relpath_cross_drive(path: str, start: str) -> str:
        if "test.py" in path:
            msg = "path is on mount 'O:', start on mount 'C:'"
            raise ValueError(msg)
        return original_relpath(path, start)

    monkeypatch.setattr(os.path, "relpath", _relpath_cross_drive)
    result = prj.run("r", "--", "test.py")
    result.assert_success()
    assert "ok" in result.out


def test_change_dir_is_created_if_not_exist(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\nchange_dir=a{/}b\ncommands=python --version"})
    result_first = prj.run("r")
    result_first.assert_success()
    assert (prj.path / "a" / "b").exists()


def test_change_dir_is_relative_to_conf(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\nchange_dir=a"})
    result = prj.run("c", "-e", "py", "-k", "change_dir", "-c", prj.path.name, from_cwd=prj.path.parent)
    result.assert_success()
    lines = result.out.splitlines()
    assert lines[1] == f"change_dir = {prj.path / 'a'}"


def test_setenv_path_not_overwritten(tox_project: ToxProjectCreator) -> None:
    cmd = "import os; print(os.environ['PATH'])"
    toml = f"""
    [env_run_base]
    package = "skip"
    set_env.PATH = "{{env:PATH}}:/custom/test/path"
    commands = [["python", "-c", "{cmd}"]]
    """
    project = tox_project({"tox.toml": toml})
    result = project.run("r")
    result.assert_success()
    # The custom path from set_env must survive — not be overwritten
    assert "/custom/test/path" in result.out


def test_setenv_path_venv_paths_first(tox_project: ToxProjectCreator) -> None:
    cmd = "import os; print(os.environ['PATH'])"
    toml = f"""
    [env_run_base]
    package = "skip"
    set_env.PATH = "{{env:PATH}}:/trailing/path"
    commands = [["python", "-c", "{cmd}"]]
    """
    project = tox_project({"tox.toml": toml})
    result = project.run("r")
    result.assert_success()
    path_line = next(line for line in result.out.splitlines() if "/trailing/path" in line)
    path_entries = path_line.split(":")
    # The virtual environment paths (containing .tox) must come before the trailing path
    tox_idx = next((i for i, p in enumerate(path_entries) if ".tox" in p), None)
    trailing_idx = next(i for i, p in enumerate(path_entries) if p == "/trailing/path")
    assert tox_idx is not None, f"expected .tox path in PATH, got: {path_line}"
    assert tox_idx < trailing_idx, f"venv paths should precede trailing path: {path_line}"


def test_cachedir_tag_created(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip"})
    result = prj.run("r")
    result.assert_success()

    tag = prj.path / ".tox" / "CACHEDIR.TAG"
    assert tag.is_file()
    content = tag.read_text(encoding="utf-8")
    assert content.startswith("Signature: 8a477f597d28d172789f06886806bc55\n")


def test_gitignore_created_in_work_dir(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands=python -c 'print(1)'"})
    result = prj.run("r")
    result.assert_success()
    gitignore = prj.path / ".tox" / ".gitignore"
    assert gitignore.exists()
    assert gitignore.read_text(encoding="utf-8") == "*\n"



from __future__ import annotations

import sys
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from tox.config.loader.replacer import MatchError

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


def test_config_in_toml_core(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
    [tool.tox]
    env_list = [ "A", "B"]

    [tool.tox.env_run_base]
    description = "Do magical things"
    commands = [
        ["python", "--version"],
        ["python", "-c", "import sys; print(sys.executable)"]
    ]
    """
    })

    outcome = project.run("c", "--core")
    outcome.assert_success()
    assert "# Exception: " not in outcome.out, outcome.out
    assert "# !!! unused: " not in outcome.out, outcome.out


def test_config_in_toml_non_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
    [tool.tox.env.C]
    description = "Do magical things in C"
    commands = [
        ["python", "--version"]
    ]
    """
    })

    outcome = project.run("c", "-e", "C", "--core")
    outcome.assert_success()
    assert "# Exception: " not in outcome.out, outcome.out
    assert "# !!! unused: " not in outcome.out, outcome.out


def test_config_in_toml_extra(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
    [tool.tox.env_run_base]
    description = "Do magical things"
    commands = [
        ["python", "--version"]
    ]
    """
    })

    outcome = project.run("c", "-e", ".".join(str(i) for i in sys.version_info[0:2]))
    outcome.assert_success()
    assert "# Exception: " not in outcome.out, outcome.out
    assert "# !!! unused: " not in outcome.out, outcome.out


def test_config_in_toml_explicit_mentioned(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
    [tool.tox.env_run_base]
    description = "Do magical things"
    commands = [
        ["python", "--version"]
    ]
    """
    })

    outcome = project.run("l", "-c", "pyproject.toml")
    outcome.assert_success()
    assert "could not recognize config file pyproject.toml" not in outcome.out, outcome.out


def test_config_in_toml_replace_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"pyproject.toml": '[tool.tox.env_run_base]\ndescription = "{missing:miss}"'})
    outcome = project.run("c", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndescription = miss\n", "")


def test_config_in_toml_replace_env_name_via_env(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": '[tool.tox.env_run_base]\ndescription = "Magic in {env:MAGICAL:{env_name}}"'
    })
    outcome = project.run("c", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndescription = Magic in py\n", "")


def test_config_in_toml_replace_env_name_via_env_set(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("MAGICAL", "YEAH")
    project = tox_project({
        "pyproject.toml": '[tool.tox.env_run_base]\ndescription = "Magic in {env:MAGICAL:{env_name}}"'
    })
    outcome = project.run("c", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndescription = Magic in YEAH\n", "")


def test_config_in_toml_replace_from_env_section_absolute(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        description = "a"
        [tool.tox.env.B]
        description = "{[tool.tox.env.A]env_name}"
        """
    })
    outcome = project.run("c", "-e", "B", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:B]\ndescription = A\n", "")


def test_config_in_toml_replace_from_section_absolute(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.extra]
        ok = "o"
        [tool.tox.env.B]
        description = "{[tool.tox.extra]ok}"
        """
    })
    outcome = project.run("c", "-e", "B", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:B]\ndescription = o\n", "")


def test_config_in_toml_replace_from_section_absolute_nok(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        extra = []
        [tool.tox.env.B]
        description = "{[tool.tox.extra.more]ok:failed}"
        """
    })
    outcome = project.run("c", "-e", "B", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err(
        "[testenv:B]\nROOT: Failed to load key more as not dictionary []\ndescription = failed\n", ""
    )


def test_config_in_toml_replace_posargs_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        commands = [["python", { replace = "posargs", default = ["a", "b"], extend = true } ]]
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "commands")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ncommands = python a b\n", "")


def test_config_in_toml_replace_posargs_empty(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        commands = [["python", { replace = "posargs", default = ["a", "b"], extend = true } ]]
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "commands", "--")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ncommands = python\n", "")


def test_config_in_toml_replace_posargs_empty_optional(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        commands = [{ replace = "posargs", default = ["a", "b"] }, ["python"]]
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "commands", "--")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ncommands = python\n", "")


def test_config_in_toml_replace_posargs_set(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        commands = [["python", { replace = "posargs", default = ["a", "b"], extend = true } ]]
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "commands", "--", "c", "d")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ncommands = python c d\n", "")


def test_config_in_toml_replace_env_default(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        description = { replace = "env", name = "NAME", default = "OK" }
        """
    })
    monkeypatch.delenv("NAME", raising=False)

    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ndescription = OK\n", "")


def test_config_in_toml_replace_env_set(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.A]
        description = { replace = "env", name = "NAME", default = "OK" }
        """
    })
    monkeypatch.setenv("NAME", "OK2")

    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ndescription = OK2\n", "")


def test_config_in_toml_replace_ref_of(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_run_base]
        extras = ["A", "{env_name}"]
        [tool.tox.env.c]
        extras = [{ replace = "ref", of = ["tool", "tox", "env_run_base", "extras"], extend = true}, "B"]
        """
    })
    monkeypatch.setenv("NAME", "OK2")

    outcome = project.run("c", "-e", "c", "-k", "extras")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:c]\nextras =\n  a\n  b\n  c\n", "")


def test_config_in_toml_replace_ref_env(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.b]
        extras = ["{env_name}"]
        [tool.tox.env.a]
        extras = [{ replace = "ref", env = "b", "key" = "extras", extend = true }, "a"]
        """
    })
    monkeypatch.setenv("NAME", "OK2")

    outcome = project.run("c", "-e", "a", "-k", "extras")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:a]\nextras =\n  a\n  b\n", "")


def test_config_in_toml_replace_env_circular_set(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.a]
        set_env.COVERAGE_FILE = { replace = "env", name = "COVERAGE_FILE", default = "{env_name}" }
        """
    })
    monkeypatch.setenv("COVERAGE_FILE", "OK")

    outcome = project.run("c", "-e", "a", "-k", "set_env")
    outcome.assert_success()
    assert "COVERAGE_FILE=OK" in outcome.out, outcome.out


def test_config_in_toml_replace_env_circular_unset(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.a]
        set_env.COVERAGE_FILE = { replace = "env", name = "COVERAGE_FILE", default = "{env_name}" }
        """
    })
    monkeypatch.delenv("COVERAGE_FILE", raising=False)

    outcome = project.run("c", "-e", "a", "-k", "set_env")
    outcome.assert_success()
    assert "COVERAGE_FILE=a" in outcome.out, outcome.out


def test_config_in_toml_replace_fails(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.B]
        description = "{[tool.tox.extra]ok:d}"
        """
    })
    outcome = project.run("c", "-e", "B", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:B]\ndescription = d\n", "")


def test_config_in_toml_replace_from_core(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env.B]
        description = "{[tool.tox]no_package}"
        """
    })
    outcome = project.run("c", "-e", "B", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:B]\ndescription = False\n", "")


def test_config_in_toml_with_legacy(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        legacy_tox_ini = '''
            [testenv]
            description=legacy
        '''
        """
    })
    outcome = project.run("c", "-e", "py", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndescription = legacy\n", "")


def test_config_in_toml_bad_type_env_name(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        env = [1]
        """
    })
    outcome = project.run("l")
    outcome.assert_failed()
    outcome.assert_out_err("ROOT: HandledError| Environment key must be string, got 1\n", "")


def test_config_in_toml_bad_type_env(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        env = {a = 1}
        """
    })
    outcome = project.run("l")
    outcome.assert_failed()
    outcome.assert_out_err("ROOT: HandledError| tool.tox.env.a must be a table, is 'int'\n", "")


def test_config_deps(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_run_base]
        deps = ['mypy>=1', 'ruff==1']
        """
    })
    outcome = project.run("c", "-k", "deps")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndeps =\n  mypy>=1\n  ruff==1\n", "")


def test_config_deps_req(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_run_base]
        deps = ['-r requirements.txt']
        """
    })
    outcome = project.run("c", "-k", "deps")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\ndeps = -r requirements.txt\n", "")


def test_config_requires(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        requires = ['tox>=4']
        """
    })
    outcome = project.run("c", "-k", "requires", "--core")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\n\n[tox]\nrequires =\n  tox>=4\n  tox\n", "")


def test_config_set_env_ref(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_run_base]
        set_env = { A = "1", B = "2"}
        [tool.tox.env.t]
        set_env = [
            { replace = "ref", of = ["tool", "tox", "env_run_base", "set_env"]},
            { C = "3", D = "4"},
        ]
        """
    })
    outcome = project.run("c", "-et", "-k", "set_env", "--hashseed", "1")
    outcome.assert_success()
    out = (
        "[testenv:t]\n"
        "set_env =\n"
        "  A=1\n"
        "  B=2\n"
        "  C=3\n"
        "  D=4\n"
        "  PIP_DISABLE_PIP_VERSION_CHECK=1\n"
        "  PYTHONHASHSEED=1\n"
        "  PYTHONIOENCODING=utf-8\n"
    )
    outcome.assert_out_err(out, "")


def test_config_set_env_substitution_deferred(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": """
        [env_run_base]
        package = "skip"
        set_env.COVERAGE_SRC = "{env_site_packages_dir}{/}mypackage"
        """
    })
    outcome = project.run("c", "-e", "py", "-k", "set_env")
    outcome.assert_success()
    assert "COVERAGE_SRC=" in outcome.out
    assert "mypackage" in outcome.out


def test_config_env_run_base_deps_reference_with_additional_deps(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_run_base]
        deps = ["pytest>=8", "coverage>=7"]

        [tool.tox.env.test]
        deps = ["{[tool.tox.env_run_base]deps}", "pytest-xdist", "pytest-timeout"]
        """
    })
    outcome = project.run("c", "-e", "test", "-k", "deps")
    outcome.assert_success()
    out = "[testenv:test]\ndeps =\n  pytest>=8\n  coverage>=7\n  pytest-xdist\n  pytest-timeout\n"
    outcome.assert_out_err(out, "")


def test_config_env_pkg_base_deps_reference_with_additional_deps(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox.env_pkg_base]
        deps = ["build", "wheel"]

        [tool.tox.env.pkg]
        deps = ["{[tool.tox.env_pkg_base]deps}", "setuptools>=40"]
        """
    })
    outcome = project.run("c", "-e", "pkg", "-k", "deps")
    outcome.assert_success()
    out = "[testenv:pkg]\ndeps =\n  build\n  wheel\n  setuptools>=40\n"
    outcome.assert_out_err(out, "")


def test_config_env_base_inherit_from_arbitrary_section(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": """
        [tool.tox]
        env_list = ["a", "b"]

        [tool.tox.env.shared]
        description = "shared config"
        skip_install = true

        [tool.tox.env.a]
        base = ["shared"]

        [tool.tox.env.b]
        base = ["shared"]
        """
    })
    outcome = project.run("c", "-e", "a,b", "-k", "description")
    outcome.assert_success()
    out = "[testenv:a]\ndescription = shared config\n\n[testenv:b]\ndescription = shared config\n"
    outcome.assert_out_err(out, "")


def test_config_in_toml_replace_glob_match(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    (tmp_path / "p" / "dist").mkdir(parents=True)
    (tmp_path / "p" / "dist" / "pkg-1.0.whl").touch()
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env.A]
        description = { replace = "glob", pattern = "dist/*.whl" }
        """),
    })
    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    assert "pkg-1.0.whl" in outcome.out


def test_config_in_toml_replace_glob_no_match_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env.A]
        description = { replace = "glob", pattern = "dist/*.xyz", default = "none" }
        """),
    })
    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ndescription = none\n", "")


def test_config_in_toml_replace_glob_extend(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    (tmp_path / "p" / "dist").mkdir(parents=True)
    (tmp_path / "p" / "dist" / "a.whl").touch()
    (tmp_path / "p" / "dist" / "b.whl").touch()
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env.A]
        commands = [["echo", { replace = "glob", pattern = "dist/*.whl", extend = true }]]
        """),
    })
    outcome = project.run("c", "-e", "A", "-k", "commands")
    outcome.assert_success()
    assert "a.whl" in outcome.out
    assert "b.whl" in outcome.out


@pytest.mark.parametrize(
    ("env_vars", "condition", "then", "else_val", "expected"),
    [
        pytest.param({"TAG": "v1"}, "env.TAG", "yes", "no", "yes", id="env_set"),
        pytest.param({}, "env.TAG", "yes", "no", "no", id="env_unset"),
        pytest.param({"TAG": ""}, "env.TAG", "yes", "no", "no", id="env_empty"),
        pytest.param({"CI": "true"}, "env.CI == 'true'", "ci", "local", "ci", id="eq_match"),
        pytest.param({"CI": "false"}, "env.CI == 'true'", "ci", "local", "local", id="eq_no_match"),
        pytest.param({"M": "s"}, "env.M != 'prod'", "dev", "prod", "dev", id="neq"),
        pytest.param({"CI": "1", "D": "1"}, "env.CI and env.D", "y", "n", "y", id="and_true"),
        pytest.param({"CI": "1"}, "env.CI and env.D", "y", "n", "n", id="and_partial"),
        pytest.param({}, "env.CI or env.L", "y", "n", "n", id="or_false"),
        pytest.param({"L": "1"}, "env.CI or env.L", "y", "n", "y", id="or_true"),
        pytest.param({}, "not env.CI", "local", "ci", "local", id="not_true"),
        pytest.param({"CI": "1"}, "not env.CI", "local", "ci", "ci", id="not_false"),
    ],
)
def test_config_in_toml_replace_if(  # noqa: PLR0913
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    env_vars: dict[str, str],
    condition: str,
    then: str,
    else_val: str,
    expected: str,
) -> None:
    for k in ("TAG", "CI", "D", "L", "M"):
        monkeypatch.delenv(k, raising=False)
    for k, v in env_vars.items():
        monkeypatch.setenv(k, v)
    project = tox_project({
        "pyproject.toml": dedent(f"""
        [tool.tox.env.A]
        description = {{ replace = "if", condition = "{condition}", then = "{then}", else = "{else_val}" }}
        """),
    })
    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err(f"[testenv:A]\ndescription = {expected}\n", "")


def test_config_in_toml_replace_if_no_else(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("DEPLOY", raising=False)
    project = tox_project({
        "pyproject.toml": """\
        [tool.tox.env.A]
        description = { replace = "if", condition = "env.DEPLOY", then = "deploy mode" }
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ndescription = \n", "")


def test_config_in_toml_replace_if_nested_substitution(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("DEPLOY", "yes")
    project = tox_project({
        "pyproject.toml": """\
        [tool.tox.env.A]
        description = { replace = "if", condition = "env.DEPLOY", then = "{env_name}", else = "none" }
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:A]\ndescription = A\n", "")


def test_config_in_toml_replace_if_set_env(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("TAG_NAME", "v2.0")
    project = tox_project({
        "pyproject.toml": """\
        [tool.tox.env.A]
        set_env.MATURITY = { replace = "if", condition = "env.TAG_NAME", then = "production", else = "testing" }
        """
    })
    outcome = project.run("c", "-e", "A", "-k", "set_env")
    outcome.assert_success()
    assert "MATURITY=production" in outcome.out


def test_config_in_toml_replace_if_extend(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("V", "1")
    toml = """\
    [tool.tox.env.A]
    commands = [["echo", { replace = "if", condition = "env.V", then = ["-v"], else = ["-q"], extend = true }]]
    """
    project = tox_project({"pyproject.toml": toml})
    outcome = project.run("c", "-e", "A", "-k", "commands")
    outcome.assert_success()
    assert "-v" in outcome.out


@pytest.mark.parametrize(
    ("condition_toml", "error_match"),
    [
        pytest.param(
            'then = "yes", else = "no"',
            "No condition was supplied in if replacement",
            id="missing_condition",
        ),
        pytest.param(
            'condition = "env.CI", else = "no"',
            "No 'then' value was supplied in if replacement",
            id="missing_then",
        ),
        pytest.param(
            'condition = "env.CI ===", then = "yes"',
            r"Invalid condition expression: env\.CI ===",
            id="invalid_syntax",
        ),
        pytest.param(
            'condition = "env.CI > env.X", then = "yes"',
            r"Unsupported comparison operator in condition: env\.CI > env\.X",
            id="unsupported_compare",
        ),
        pytest.param(
            'condition = "1 + 2", then = "yes"',
            r"Unsupported expression in condition: ",
            id="unsupported_expr",
        ),
    ],
)
def test_config_in_toml_replace_if_error(tox_project: ToxProjectCreator, condition_toml: str, error_match: str) -> None:
    project = tox_project({
        "pyproject.toml": dedent(f"""
        [tool.tox.env.A]
        description = {{ replace = "if", {condition_toml} }}
        """),
    })
    with pytest.raises(MatchError, match=error_match):
        project.run("c", "-e", "A", "-k", "description")


def test_config_in_toml_replace_if_factor_positive(tox_project: ToxProjectCreator) -> None:

    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env."3.13-django50"]
        description = { replace = "if", condition = "factor.django50", then = "has django50", else = "no django50" }
        """),
    })
    outcome = project.run("c", "-e", "3.13-django50", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:3.13-django50]\ndescription = has django50\n", "")


def test_config_in_toml_replace_if_factor_negative(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env."3.13"]
        description = { replace = "if", condition = "factor.django50", then = "has django50", else = "no django50" }
        """),
    })
    outcome = project.run("c", "-e", "3.13", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:3.13]\ndescription = no django50\n", "")


def test_config_in_toml_replace_if_factor_platform(tox_project: ToxProjectCreator) -> None:
    condition_val = f"factor.{sys.platform}"
    toml_str = f"""
        [tool.tox.env.task]
        description.replace = "if"
        description.condition = "{condition_val}"
        description.then = "correct platform"
        description.else = "wrong platform"
        """
    project = tox_project({"pyproject.toml": dedent(toml_str)})
    outcome = project.run("c", "-e", "task", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:task]\ndescription = correct platform\n", "")


def test_config_in_toml_replace_if_factor_not(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env."3.13"]
        description = { replace = "if", condition = "not factor.win32", then = "not windows", else = "windows" }
        """),
    })
    outcome = project.run("c", "-e", "3.13", "-k", "description")
    outcome.assert_success()
    expected = "windows" if sys.platform == "win32" else "not windows"
    outcome.assert_out_err(f"[testenv:3.13]\ndescription = {expected}\n", "")


def test_config_in_toml_replace_if_factor_and(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env.py313-django50]
        description.replace = "if"
        description.condition = "factor.django50 and factor.py313"
        description.then = "both"
        description.else = "not both"
        """),
    })
    outcome = project.run("c", "-e", "py313-django50", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py313-django50]\ndescription = both\n", "")


def test_config_in_toml_replace_if_factor_or(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env.py313]
        description.replace = "if"
        description.condition = "factor.django50 or factor.py313"
        description.then = "at least one"
        description.else = "neither"
        """),
    })
    outcome = project.run("c", "-e", "py313", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py313]\ndescription = at least one\n", "")


def test_config_in_toml_replace_if_factor_combined_with_env(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("CI", "1")
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env."3.13-django50"]
        description.replace = "if"
        description.condition = "factor.django50 and env.CI"
        description.then = "django in CI"
        description.else = "other"
        """),
    })
    outcome = project.run("c", "-e", "3.13-django50", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:3.13-django50]\ndescription = django in CI\n", "")


def test_config_in_toml_replace_if_list_without_extend_in_deps(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("X", "1")
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env_run_base]
        deps = [
            "pytest",
            { replace = "if", condition = "env.X", then = ["extra-pkg"] },
        ]
        """),
    })
    outcome = project.run("c", "-e", "py", "-k", "deps")
    outcome.assert_failed()
    assert "failed to load py.deps: deps expected str, list[str], or list[Requirement]" in outcome.out


def test_config_in_toml_replace_if_list_with_extend(
    tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("X", "1")
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env_run_base]
        deps = [
            "pytest",
            { replace = "if", condition = "env.X", then = ["extra-pkg"], extend = true },
        ]
        """),
    })
    outcome = project.run("c", "-e", "py", "-k", "deps")
    outcome.assert_success()
    assert "extra-pkg" in outcome.out


_CMD = 'commands = [["python", "--version"]]'


@pytest.mark.parametrize(
    ("toml_body", "key", "expected_msg"),
    [
        # --- str fields ---
        pytest.param(f"description = 1\n{_CMD}", "description", "1 is not of type 'str'", id="description-int"),
        pytest.param(f"platform = 1\n{_CMD}", "platform", "1 is not of type 'str'", id="platform-int"),
        pytest.param(f"runner = 1\n{_CMD}", "runner", "1 is not of type 'str'", id="runner-int"),
        pytest.param(f"package = 1\n{_CMD}", "package", "1 is not of type 'str'", id="package-int"),
        pytest.param(f"pylock = 1\n{_CMD}", "pylock", "1 is not of type 'str'", id="pylock-int"),
        pytest.param(f"virtualenv_spec = 1\n{_CMD}", "virtualenv_spec", "1 is not of type 'str'", id="venv-spec-int"),
        # --- bool fields ---
        pytest.param(f"ignore_errors = 42\n{_CMD}", "ignore_errors", "42 is not of type 'bool'", id="ignore-err-int"),
        pytest.param(
            f"parallel_show_output = 42\n{_CMD}", "parallel_show_output", "42 is not of type 'bool'", id="parallel-int"
        ),
        pytest.param(f"recreate = 42\n{_CMD}", "recreate", "42 is not of type 'bool'", id="recreate-int"),
        pytest.param(
            f"args_are_paths = 42\n{_CMD}", "args_are_paths", "42 is not of type 'bool'", id="args-are-paths-int"
        ),
        pytest.param(
            f"ignore_outcome = 42\n{_CMD}", "ignore_outcome", "42 is not of type 'bool'", id="ignore-outcome-int"
        ),
        pytest.param(f"fail_fast = 42\n{_CMD}", "fail_fast", "42 is not of type 'bool'", id="fail-fast-int"),
        pytest.param(f"skip_install = 42\n{_CMD}", "skip_install", "42 is not of type 'bool'", id="skip-install-int"),
        pytest.param(f"use_develop = 42\n{_CMD}", "use_develop", "42 is not of type 'bool'", id="use-develop-int"),
        pytest.param(
            f"system_site_packages = 42\n{_CMD}",
            "system_site_packages",
            "42 is not of type 'bool'",
            id="sys-site-pkg-int",
        ),
        pytest.param(f"always_copy = 42\n{_CMD}", "always_copy", "42 is not of type 'bool'", id="always-copy-int"),
        pytest.param(f"download = 42\n{_CMD}", "download", "42 is not of type 'bool'", id="download-int"),
        pytest.param(f"pip_pre = 42\n{_CMD}", "pip_pre", "42 is not of type 'bool'", id="pip-pre-int"),
        pytest.param(
            f"constrain_package_deps = 42\n{_CMD}",
            "constrain_package_deps",
            "42 is not of type 'bool'",
            id="constrain-pkg-deps-int",
        ),
        pytest.param(
            f"use_frozen_constraints = 42\n{_CMD}",
            "use_frozen_constraints",
            "42 is not of type 'bool'",
            id="frozen-constraints-int",
        ),
        pytest.param(
            f"skip_missing_interpreters = 42\n{_CMD}",
            "skip_missing_interpreters",
            "42 is not of type 'bool'",
            id="skip-missing-int",
        ),
        # --- int fields ---
        pytest.param(
            f'commands_retry = "bad"\n{_CMD}',
            "commands_retry",
            "invalid literal for int() with base 10: 'bad'",
            id="cmd-retry-str",
        ),
        # --- float fields ---
        pytest.param(
            f'suicide_timeout = "bad"\n{_CMD}',
            "suicide_timeout",
            "could not convert string to float: 'bad'",
            id="suicide-str",
        ),
        pytest.param(
            f'interrupt_timeout = "bad"\n{_CMD}',
            "interrupt_timeout",
            "could not convert string to float: 'bad'",
            id="interrupt-str",
        ),
        pytest.param(
            f'terminate_timeout = "bad"\n{_CMD}',
            "terminate_timeout",
            "could not convert string to float: 'bad'",
            id="terminate-str",
        ),
        # --- Path fields ---
        pytest.param(f"env_dir = 1\n{_CMD}", "env_dir", "1 is not of type 'str'", id="env-dir-int"),
        pytest.param(f"env_tmp_dir = 1\n{_CMD}", "env_tmp_dir", "1 is not of type 'str'", id="env-tmp-dir-int"),
        pytest.param(f"env_log_dir = 1\n{_CMD}", "env_log_dir", "1 is not of type 'str'", id="env-log-dir-int"),
        pytest.param(f"change_dir = 1\n{_CMD}", "change_dir", "1 is not of type 'str'", id="change-dir-int"),
        # --- list[str] fields ---
        pytest.param(f"pass_env = 42\n{_CMD}", "pass_env", "42 is not list", id="pass-env-not-list"),
        pytest.param(f"pass_env = [1]\n{_CMD}", "pass_env", "1 is not of type 'str'", id="pass-env-int-item"),
        pytest.param(
            f"disallow_pass_env = 42\n{_CMD}", "disallow_pass_env", "42 is not list", id="disallow-pass-env-not-list"
        ),
        pytest.param(
            f"allowlist_externals = [true]\n{_CMD}",
            "allowlist_externals",
            "True is not of type 'str'",
            id="allowlist-bool-item",
        ),
        pytest.param(
            f"base_python = 42\n{_CMD}",
            "base_python",
            "42 cannot cast to list[str] | str",
            id="base-python-not-list",
        ),
        # --- set[str] fields ---
        pytest.param(f"labels = 42\n{_CMD}", "labels", "42 is not list", id="labels-not-list"),
        pytest.param(f"labels = [1]\n{_CMD}", "labels", "1 is not of type 'str'", id="labels-int-item"),
        pytest.param(f"extras = 42\n{_CMD}", "extras", "42 is not list", id="extras-not-list"),
        pytest.param(
            f"dependency_groups = 42\n{_CMD}", "dependency_groups", "42 is not list", id="dep-groups-not-list"
        ),
        # --- list[Command] fields ---
        pytest.param("commands = 42", "commands", "42 is not list", id="commands-not-list"),
        pytest.param("commands = [[1]]", "commands", "1 is not of type 'str'", id="commands-int-arg"),
        pytest.param("commands_pre = 42", "commands_pre", "42 is not list", id="commands-pre-not-list"),
        pytest.param("commands_post = 42", "commands_post", "42 is not list", id="commands-post-not-list"),
        pytest.param(
            "recreate_commands = [[1]]", "recreate_commands", "1 is not of type 'str'", id="recreate-cmds-int-arg"
        ),
        pytest.param(
            "extra_setup_commands = 42", "extra_setup_commands", "42 is not list", id="extra-setup-cmds-not-list"
        ),
        # --- EnvList fields ---
        pytest.param(f"depends = 42\n{_CMD}", "depends", "env_list must be a list, got int", id="depends-not-list"),
        pytest.param(
            f"depends = [1]\n{_CMD}",
            "depends",
            "env_list items must be strings or product dicts, got int",
            id="depends-int-item",
        ),
        # --- factory fields ---
        pytest.param("deps = [1]", "deps", "deps expected str, list[str], or list[Requirement]", id="deps-int-item"),
        pytest.param("deps = true", "deps", "deps expected str, list[str], or list[Requirement]", id="deps-bool"),
        pytest.param(
            f"constraints = 42\n{_CMD}",
            "constraints",
            "constraints expected str, list[str], or list[Requirement]",
            id="constraints-int",
        ),
    ],
)
def test_config_in_toml_type_error_message(
    tox_project: ToxProjectCreator,
    toml_body: str,
    key: str,
    expected_msg: str,
) -> None:
    project = tox_project({
        "pyproject.toml": dedent(f"""
        [tool.tox.env_run_base]
        {toml_body}
        """),
    })
    outcome = project.run("c", "-e", "py", "-k", key)
    outcome.assert_failed()
    assert f"failed to load py.{key}" in outcome.out
    assert expected_msg in outcome.out


def test_config_in_toml_handled_error_on_run(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": dedent("""
        [tool.tox.env_run_base]
        deps = [1]
        commands = [["python", "--version"]]
        """),
    })
    outcome = project.run("r", "-e", "py")
    outcome.assert_failed()
    assert "internal error" not in outcome.out
    assert "failed to load py.deps" in outcome.out
    assert "deps expected str, list[str], or list[Requirement]" in outcome.out



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_unused_config_testenv_warns_at_verbose(tox_project: ToxProjectCreator) -> None:
    toml = '[env_run_base]\npackage = "skip"\ncommands = [["python", "-c", ""]]\nbogus_key = "yes"\n'
    outcome = tox_project({"tox.toml": toml}).run("r", "-v")
    outcome.assert_success()
    assert "[testenv:py] unused config key(s): bogus_key" in outcome.out


def test_unused_config_core_warns_at_verbose(tox_project: ToxProjectCreator) -> None:
    toml = 'not_a_real_key = "yes"\n[env_run_base]\npackage = "skip"\ncommands = [["python", "-c", ""]]\n'
    outcome = tox_project({"tox.toml": toml}).run("r", "-v")
    outcome.assert_success()
    assert "[tox] unused config key(s): not_a_real_key" in outcome.out


def test_unused_config_no_warning_at_default_verbosity(tox_project: ToxProjectCreator) -> None:
    toml = '[env_run_base]\npackage = "skip"\ncommands = [["python", "-c", ""]]\nbogus_key = "yes"\n'
    outcome = tox_project({"tox.toml": toml}).run("r")
    outcome.assert_success()
    assert "unused config key" not in outcome.out


def test_no_unused_config_warning_when_all_valid(tox_project: ToxProjectCreator) -> None:
    toml = '[env_run_base]\npackage = "skip"\ncommands = [["python", "-c", ""]]\n'
    outcome = tox_project({"tox.toml": toml}).run("r", "-v")
    outcome.assert_success()
    assert "unused config key" not in outcome.out



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

import pytest

from tox.config.loader.ini.factor import LATEST_PYTHON_MINOR_MAX, LATEST_PYTHON_MINOR_MIN
from tox.config.loader.toml._product import (  # noqa: PLC2701
    _expand_range,
    expand_factor_group,
    expand_product,
)

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_expand_product_two_groups() -> None:
    result = expand_product({"product": [["a", "b"], ["x", "y"]]})
    assert result == ["a-x", "a-y", "b-x", "b-y"]


def test_expand_product_three_groups() -> None:
    result = expand_product({"product": [["a"], ["b", "c"], ["d"]]})
    assert result == ["a-b-d", "a-c-d"]


def test_expand_product_single_group() -> None:
    result = expand_product({"product": [["a", "b"]]})
    assert result == ["a", "b"]


def test_expand_product_empty() -> None:
    assert expand_product({"product": []}) == []


def test_expand_product_with_exclusion() -> None:
    result = expand_product({"product": [["a", "b"], ["x", "y"]], "exclude": ["a-y", "b-x"]})
    assert result == ["a-x", "b-y"]


def test_expand_product_exclusion_miss_is_ignored() -> None:
    result = expand_product({"product": [["a"], ["x"]], "exclude": ["nonexistent"]})
    assert result == ["a-x"]


def test_expand_product_not_list() -> None:
    with pytest.raises(TypeError, match="product value must be a list"):
        expand_product({"product": "bad"})


def test_expand_factor_group_list() -> None:
    assert expand_factor_group(["py312", "py313"]) == ["py312", "py313"]


def test_expand_factor_group_range_dict() -> None:
    assert expand_factor_group({"prefix": "py3", "start": 12, "stop": 14}) == ["py312", "py313", "py314"]


def test_expand_factor_group_bad_type() -> None:
    with pytest.raises(TypeError, match="factor group must be a list of strings or a range dict"):
        expand_factor_group(42)


def test_expand_factor_group_dict_no_prefix() -> None:
    with pytest.raises(TypeError, match="factor group must be a list of strings or a range dict"):
        expand_factor_group({"start": 1, "stop": 3})


def test_expand_range_closed() -> None:
    assert _expand_range({"prefix": "py3", "start": 12, "stop": 14}) == ["py312", "py313", "py314"]


def test_expand_range_open_stop() -> None:
    result = _expand_range({"prefix": "py3", "start": 12})
    assert result == [f"py3{i}" for i in range(12, LATEST_PYTHON_MINOR_MAX + 1)]


def test_expand_range_open_start() -> None:
    result = _expand_range({"prefix": "py3", "stop": 13})
    assert result == [f"py3{i}" for i in range(LATEST_PYTHON_MINOR_MIN, 14)]


def test_expand_range_no_bounds() -> None:
    with pytest.raises(TypeError, match="range must have at least 'start' or 'stop'"):
        _expand_range({"prefix": "py3"})


def test_expand_range_start_not_int() -> None:
    with pytest.raises(TypeError, match="range 'start' must be an integer"):
        _expand_range({"prefix": "py3", "start": "12", "stop": 14})


def test_expand_range_stop_not_int() -> None:
    with pytest.raises(TypeError, match="range 'stop' must be an integer"):
        _expand_range({"prefix": "py3", "start": 12, "stop": "14"})


def test_expand_product_mixed_list_and_range() -> None:
    result = expand_product({
        "product": [
            {"prefix": "py3", "start": 12, "stop": 13},
            ["django42", "django50"],
        ],
    })
    assert result == ["py312-django42", "py312-django50", "py313-django42", "py313-django50"]


def test_product_envs_listed(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                { product = [["py312", "py313"], ["django42", "django50"]] },
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    for env in ("py312-django42", "py312-django50", "py313-django42", "py313-django50"):
        assert env in result.out


def test_product_mixed_with_literals(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                "lint",
                { product = [["py312", "py313"], ["django42"]] },
                "docs",
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    for env in ("lint", "py312-django42", "py313-django42", "docs"):
        assert env in result.out


def test_product_with_range(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                { product = [{ prefix = "py3", start = 12, stop = 13 }, ["django42"]] },
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    assert "py312-django42" in result.out
    assert "py313-django42" in result.out


def test_product_with_exclusion(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                { product = [["py312", "py313"], ["django42", "django50"]], exclude = ["py312-django50"] },
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    assert "py312-django42" in result.out
    assert "py313-django42" in result.out
    assert "py313-django50" in result.out
    assert "py312-django50" not in result.out


def test_product_multiple_in_env_list(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                { product = [["py312"], ["django42"]] },
                { product = [["py313"], ["flask20"]] },
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    assert "py312-django42" in result.out
    assert "py313-flask20" in result.out


def test_product_deduplication(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            env_list = [
                "py312-django42",
                { product = [["py312"], ["django42"]] },
            ]

            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("l")
    result.assert_success()
    assert result.out.count("py312-django42") == 1



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_conf_in_setup_cfg(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"setup.cfg": "[tox:tox]\nenv_list=\n a\n b"})

    outcome = project.run("l")
    outcome.assert_success()
    assert outcome.out == "default environments:\na -> [no description]\nb -> [no description]\n"


def test_setup_cfg_without_tox_section(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"setup.cfg": "[tox]\nenv_list=\n a\n b"})
    filename = str(project.path / "setup.cfg")
    outcome = project.run("l", "-c", filename)
    outcome.assert_failed()
    assert outcome.out == f"ROOT: HandledError| could not recognize config file {filename}\n"


def test_setup_cfg_non_tox_section_not_discovered_as_env(tox_project: ToxProjectCreator) -> None:
    cfg = "[tox:tox]\nenv_list = py\n[tox:testenv]\npackage = skip\n[options]\npackages = find:\n"
    project = tox_project({"setup.cfg": cfg})
    outcome = project.run("l")
    outcome.assert_success()
    assert "find" not in outcome.out


def test_setup_cfg_does_not_block_pyproject_toml(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "setup.cfg": "[metadata]\nname = test-pkg\n",
        "pyproject.toml": "[tool.tox]\nenv_list = ['a']\n\n[tool.tox.env_run_base]\npackage = 'skip'\n",
    })
    outcome = project.run("l")
    outcome.assert_success()
    assert "a -> [no description]" in outcome.out


def test_setup_cfg_does_not_block_tox_toml(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "setup.cfg": "[metadata]\nname = test-pkg\n",
        "tox.toml": 'env_list = ["b"]\n\n[env_run_base]\npackage = "skip"\n',
    })
    outcome = project.run("l")
    outcome.assert_success()
    assert "b -> [no description]" in outcome.out


def test_setup_cfg_unicode_characters(tox_project: ToxProjectCreator) -> None:
    """Test that setup.cfg files with unicode characters can be read without UnicodeDecodeError."""
    cfg = """
        [tox:tox]
        env_list=
         a

        [tox:testenv:a]
        description = Test with emoji ❌ and unicode ✨
        package = skip
        """
    project = tox_project({"setup.cfg": dedent(cfg)})
    outcome = project.run("l")
    outcome.assert_success()
    assert "a ->" in outcome.out



from __future__ import annotations

import sys
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

from packaging.version import Version

from tox.plugin import impl
from tox.version import version as __version__

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    from tox.config.cli.parser import CORE  # noqa: PLC0415

    our = parser.add_command(
        "quickstart",
        ["q"],
        "Command line script to quickly create a tox config file for a Python project",
        quickstart,
        inherit=frozenset({CORE}),
    )
    our.add_argument(
        "quickstart_root",
        metavar="root",
        default=Path().absolute(),
        nargs="?",
        help="folder to create the tox.ini file",
        type=Path,
    )


def quickstart(state: State) -> int:
    root = state.conf.options.quickstart_root.absolute()
    tox_ini = root / "tox.ini"
    if tox_ini.exists():
        print(f"{tox_ini} already exist, refusing to overwrite")  # noqa: T201
        return 1
    version = str(Version(__version__.split("+")[0]))
    text = f"""
        [tox]
        env_list =
            py{"".join(str(i) for i in sys.version_info[0:2])}
        minversion = {version}

        [testenv]
        description = run the tests with pytest
        package = wheel
        wheel_build_env = .pkg
        deps =
            pytest>=6
        commands =
            pytest {{tty:--color=yes}} {{posargs}}
    """
    content = dedent(text).lstrip()

    print(f"tox {__version__} quickstart utility, will create {tox_ini}:")  # noqa: T201
    print(content, end="")  # noqa: T201

    root.mkdir(parents=True, exist_ok=True)
    tox_ini.write_text(content)
    return 0



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

from tox.config.loader.section import Section
from tox.config.sets import ConfigSet, CoreConfigSet
from tox.config.source.ini import IniSource

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture

    from tests.conftest import ToxIniCreator


def test_source_ini_with_interpolated(tmp_path: Path) -> None:
    loader = IniSource(tmp_path, content="[tox]\na = %(c)s").get_loader(Section(None, "tox"), {})
    assert loader is not None
    loader.load_raw("a", None, None)


def test_source_ini_ignore_non_testenv_sections(tmp_path: Path, mocker: MockerFixture) -> None:
    core = mocker.create_autospec(CoreConfigSet, instance=True)
    core.__getitem__ = lambda _self, key: [] if key == "env_list" else None
    loader = IniSource(tmp_path, content="[mypy-rest_framework.compat.*]")
    res = list(loader.envs(core))
    assert not res


def test_source_ini_ignore_invalid_factor_filters(tmp_path: Path, mocker: MockerFixture) -> None:
    core = mocker.create_autospec(CoreConfigSet, instance=True)
    core.__getitem__ = lambda _self, key: [] if key == "env_list" else None
    loader = IniSource(tmp_path, content="[a]\nb= if c: d")
    res = list(loader.envs(core))
    assert not res


def test_source_ini_custom_non_testenv_sections(tox_ini_conf: ToxIniCreator) -> None:
    """Validate that a plugin can load section with custom prefix overlapping testenv name."""

    class CustomConfigSet(ConfigSet):
        def register_config(self) -> None:
            self.add_config(
                keys=["a"],
                of_type=str,
                default="",
                desc="d",
            )

    config = tox_ini_conf("[testenv:foo]\n[custom:foo]\na = b")
    known_envs = list(config._src.envs(config.core))  # noqa: SLF001
    assert known_envs
    custom_section = config.get_section_config(
        section=Section("custom", "foo"),
        base=[],
        of_type=CustomConfigSet,
        for_env=None,
    )
    assert custom_section["a"] == "b"


def test_source_ini_unicode_characters(tmp_path: Path) -> None:
    """Test that INI files with unicode characters are read correctly."""
    ini_file = tmp_path / "tox.ini"
    content = """
        [tox]
        description = Test with emoji ❌ and unicode ✨
        """
    ini_file.write_text(dedent(content), encoding="utf-8")
    loader = IniSource(ini_file).get_loader(Section(None, "tox"), {})
    assert loader is not None
    desc = loader.load_raw("description", None, None)
    assert desc == "Test with emoji ❌ and unicode ✨"



"""Declare the abstract base class for tox environments that handle the Python language."""

from __future__ import annotations

import logging
import re
import sys
import sysconfig
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

from virtualenv.discovery.py_spec import PythonSpec

from tox.tox_env.api import ToxEnv, ToxEnvCreateArgs
from tox.tox_env.errors import Fail, Recreate, Skip

if TYPE_CHECKING:
    from tox.config.main import Config


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


@dataclass(frozen=True)
class PythonInfo:
    implementation: str
    version_info: VersionInfo
    version: str
    is_64: bool
    platform: str
    extra: dict[str, Any]
    free_threaded: bool = False
    machine: str | None = None

    @property
    def version_no_dot(self) -> str:
        return f"{self.version_info.major}{self.version_info.minor}"

    @property
    def impl_lower(self) -> str:
        return self.implementation.lower()

    @property
    def version_dot(self) -> str:
        return f"{self.version_info.major}.{self.version_info.minor}"


PY_FACTORS_RE = re.compile(
    r"""
    ^(?!py$)                                               # don't match 'py' as it doesn't provide any info
    (?P<impl>py|pypy|cpython|jython|graalpy|rustpython|ironpython) # the interpreter; most users will simply use 'py'
    (?:
    (?P<version>[2-9]\.?[0-9]?[0-9]?)                      # the version; one of: MAJORMINOR, MAJOR.MINOR
    (?P<threaded>t?)                                       # version followed by t for free-threading
    )?$
    """,
    re.VERBOSE,
)
PY_FACTORS_RE_EXPLICIT_VERSION = re.compile(
    r"""
    ^
    ( (?P<impl> cpython | pypy ) - )?   # optional interpreter prefix with dash
    (?P<version> [23] \. [0-9]+ )        # explicit major.minor version (Python 2.x or 3.x)
    (?P<threaded> t? )                   # optional free-threaded suffix
    $
    """,
    re.VERBOSE,
)


class Python(ToxEnv, ABC):
    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        self._base_python: PythonInfo | None = None
        self._base_python_searched: bool = False
        super().__init__(create_args)

    def register_config(self) -> None:
        super().register_config()

        def _ensure_list(value: list[str] | str) -> list[str]:
            return [value] if isinstance(value, str) else value

        self.conf.add_config(  # ty: ignore[no-matching-overload] # https://github.com/astral-sh/ty/issues/2428
            keys=["default_base_python"],
            of_type=list[str] | str,
            default=[sys.executable],
            desc="fallback python interpreter used when no factor or explicit base_python is defined",
            post_process=_ensure_list,
        )

        def validate_base_python(value: list[str] | str) -> list[str]:
            return self._validate_base_python(self.name, _ensure_list(value), self.core["ignore_base_python_conflict"])

        self.conf.add_config(  # ty: ignore[no-matching-overload] # https://github.com/astral-sh/ty/issues/2428
            keys=["base_python", "basepython"],
            of_type=list[str] | str,
            default=self._base_python_default,
            desc="environment identifier for python, first one found wins",
            post_process=validate_base_python,
        )
        self.core.add_config(
            keys=["ignore_base_python_conflict", "ignore_basepython_conflict"],
            of_type=bool,
            default=False,
            desc="do not raise error if the environment name conflicts with base python",
        )
        self.conf.add_constant(
            keys=["env_site_packages_dir", "envsitepackagesdir"],
            desc="the python environments site package - pure python (purelib)",
            value=self.env_site_package_dir,
        )
        self.conf.add_constant(
            keys=["env_site_packages_dir_plat", "envsitepackagesdir_plat"],
            desc="the python environments platform-specific site package (platlib)",
            value=self.env_site_package_dir_plat,
        )
        self.conf.add_constant(
            keys=["env_bin_dir", "envbindir"],
            desc="the python environments binary folder",
            value=self.env_bin_dir,
        )
        self.conf.add_constant(
            ["env_python", "envpython"],
            desc="python executable from within the tox environment",
            value=self.env_python,
        )
        self.conf.add_constant("py_dot_ver", "<python major>.<python minor>", value=self.py_dot_ver)
        self.conf.add_constant("py_impl", "python implementation", value=self.py_impl)
        self.conf.add_constant("py_free_threaded", "is no-gil interpreted", value=self.py_free_threaded)

    def _default_set_env(self) -> dict[str, str]:
        env = super()._default_set_env()
        hash_seed: int | None = getattr(self.options, "hash_seed", None)
        if hash_seed is not None:
            env["PYTHONHASHSEED"] = str(hash_seed)
        return env

    def py_dot_ver(self) -> str:
        return self.base_python.version_dot

    def py_free_threaded(self) -> bool:
        return self.base_python.free_threaded

    def py_impl(self) -> str:
        return self.base_python.impl_lower

    def _default_pass_env(self) -> list[str]:
        env = super()._default_pass_env()
        if sys.platform == "win32":  # pragma: win32 cover
            env.extend(
                [
                    "APPDATA",  # Needed for PIP platformsdirs.windows
                    "LOCALAPPDATA",  # Needed for pymanager
                    "PROGRAMDATA",  # needed for discovering the VS compiler
                    "PROGRAMFILES(x86)",  # needed for discovering the VS compiler
                    "PROGRAMFILES",  # needed for discovering the VS compiler
                    "SYSTEMDRIVE",
                    "SYSTEMROOT",  # needed for python's crypto module
                    "COMSPEC",  # needed for distutils cygwin compiler
                    "PROCESSOR_ARCHITECTURE",  # platform.machine()
                    "NUMBER_OF_PROCESSORS",  # multiprocessing.cpu_count()
                ],
            )
        binary_extension_build = ["PKG_CONFIG", "PKG_CONFIG_PATH", "PKG_CONFIG_SYSROOT_DIR"]
        env.extend(binary_extension_build)  # used by binary extensions during installation
        env.extend(["REQUESTS_CA_BUNDLE"])
        return env

    def _base_python_default(self, conf: Config, env_name: str | None) -> list[str]:  # noqa: ARG002
        base_python = None if env_name is None else self.extract_base_python(env_name)
        return self.conf["default_base_python"] if base_python is None else [base_python]

    @classmethod
    def extract_base_python(cls, env_name: str) -> str | None:
        candidates: list[str] = []
        if match := PY_FACTORS_RE_EXPLICIT_VERSION.match(env_name):
            found = match.groupdict()
            candidates.append(f"{'pypy' if found['impl'] == 'pypy' else ''}{found['version']}{found['threaded']}")
        else:
            for factor in env_name.split("-"):
                if match := PY_FACTORS_RE.match(factor):
                    candidates.append(factor)
                elif match := PY_FACTORS_RE_EXPLICIT_VERSION.match(factor):
                    found = match.groupdict()
                    candidates.append(
                        f"{'pypy' if found['impl'] == 'pypy' else ''}{found['version']}{found['threaded']}"
                    )
        if candidates:
            if len(candidates) > 1:
                msg = f"conflicting factors {', '.join(candidates)} in {env_name}"
                raise ValueError(msg)
            return next(iter(candidates))
        return None

    @classmethod
    def _python_spec_for_sys_executable(cls) -> PythonSpec:
        implementation = sys.implementation.name
        version = sys.version_info
        bits = "64" if sys.maxsize > 2**32 else "32"
        threaded = "t" if sysconfig.get_config_var("Py_GIL_DISABLED") == 1 else ""
        parts = sysconfig.get_platform().rsplit("-", 1)
        machine_suffix = f"-{isa}" if len(parts) > 1 and (isa := parts[-1]) else ""
        string_spec = f"{implementation}{version.major}{version.minor}{threaded}-{bits}{machine_suffix}"
        return PythonSpec.from_string_spec(string_spec)

    @classmethod
    def _validate_base_python(
        cls,
        env_name: str,
        base_pythons: list[str],
        ignore_base_python_conflict: bool,  # noqa: FBT001
    ) -> list[str]:
        env_base_python = cls.extract_base_python(env_name)
        if env_base_python is not None:
            spec_name = PythonSpec.from_string_spec(env_base_python)
            for base_python in base_pythons:
                spec_base = PythonSpec.from_string_spec(base_python)
                if spec_base.path is not None:
                    path = Path(spec_base.path).absolute()
                    if str(spec_base.path) == sys.executable:
                        spec_base = cls._python_spec_for_sys_executable()
                    else:
                        spec_base = cls.python_spec_for_path(path)
                if any(
                    getattr(spec_base, key) != getattr(spec_name, key)
                    for key in ("implementation", "major", "minor", "micro", "architecture", "machine", "free_threaded")
                    if getattr(spec_name, key) is not None
                ):
                    msg = f"env name {env_name} conflicting with base python {base_python}"
                    if ignore_base_python_conflict:
                        # ignore the base python settings and return the thing that looks like a Python version
                        return [env_base_python]
                    raise Fail(msg)
        return base_pythons

    @classmethod
    @abstractmethod
    def python_spec_for_path(cls, path: Path) -> PythonSpec:
        """Get the spec for an absolute path to a Python executable.

        :param path: the path investigated

        :returns: the found spec

        """
        raise NotImplementedError

    @abstractmethod
    def env_site_package_dir(self) -> Path:
        """Return the pure-python site-packages directory (purelib) for this environment.

        Debian derivatives change site-packages to dist-packages, so we look at the last path under prefix.

        """
        raise NotImplementedError

    def env_site_package_dir_plat(self) -> Path:
        """Return the platform-specific site-packages directory (platlib) for this environment.

        On most platforms this is the same as purelib, but on some Linux distributions (Fedora, RHEL) platlib uses lib64
        instead of lib. Defaults to purelib so third-party runners don't need to override this.

        """
        return self.env_site_package_dir()

    @abstractmethod
    def env_python(self) -> Path:
        """The python executable within the tox environment."""
        raise NotImplementedError

    @abstractmethod
    def env_bin_dir(self) -> Path:
        """The binary folder within the tox environment."""
        raise NotImplementedError

    def _setup_env(self) -> None:
        """Setup a virtual python environment."""
        super()._setup_env()
        self.ensure_python_env()
        self._paths = self.prepend_env_var_path()  # now that the environment exist we can add them to the path

    def ensure_python_env(self) -> None:
        conf = self.python_cache()
        with self.cache.compare(conf, Python.__name__) as (eq, old):
            if old is None:  # does not exist -> create
                self.create_python_env()
            elif eq is False:  # pragma: no branch # exists but changed -> recreate
                raise Recreate(self._diff_msg(conf, old))

    @staticmethod
    def _diff_msg(conf: dict[str, Any], old: dict[str, Any]) -> str:
        result: list[str] = []
        added = [f"{k}={v!r}" for k, v in conf.items() if k not in old]
        if added:  # pragma: no branch
            result.append(f"added {' | '.join(added)}")
        removed = [f"{k}={v!r}" for k, v in old.items() if k not in conf]
        if removed:
            result.append(f"removed {' | '.join(removed)}")
        changed = [f"{k}={old[k]!r}->{v!r}" for k, v in conf.items() if k in old and v != old[k]]
        if changed:
            result.append(f"changed {' | '.join(changed)}")
        return f"python {', '.join(result)}"

    @abstractmethod
    def prepend_env_var_path(self) -> list[Path]:
        raise NotImplementedError

    def _done_with_setup(self) -> None:
        """Called when setup is done."""
        super()._done_with_setup()
        if self.journal or self.options.list_dependencies:
            outcome = self.installer.installed()
            if self.journal:
                self.journal["installed_packages"] = outcome
            if self.options.list_dependencies:
                logging.warning(",".join(outcome))

    def python_cache(self) -> dict[str, Any]:
        return {
            "version_info": list(self.base_python.version_info),
        }

    @property
    def base_python(self) -> PythonInfo:
        """Resolve base python."""
        base_pythons: list[str] = self.conf["base_python"]

        if self._base_python_searched is False:
            self._base_python_searched = True
            self._base_python = self._get_python(base_pythons)
            if self._base_python is not None and self.journal:
                value = self._get_env_journal_python()
                self.journal["python"] = value

        if self._base_python is None:
            if self.conf["skip_missing_interpreters"]:
                msg = f"could not find python interpreter with spec(s): {', '.join(base_pythons)}"
                raise Skip(msg)
            raise NoInterpreter(base_pythons)

        return self._base_python

    def _get_env_journal_python(self) -> dict[str, Any]:
        return {
            "implementation": self.base_python.implementation,
            "version_info": tuple(self.base_python.version_info),
            "version": self.base_python.version,
            "is_64": self.base_python.is_64,
            "sysplatform": self.base_python.platform,
            "extra_version_info": None,
            "free_threaded": self.base_python.free_threaded,
            "machine": self.base_python.machine,
        }

    @abstractmethod
    def _get_python(self, base_python: list[str]) -> PythonInfo | None:
        raise NotImplementedError

    @abstractmethod
    def create_python_env(self) -> None:
        raise NotImplementedError


class NoInterpreter(Fail):
    """could not find interpreter."""

    def __init__(self, base_pythons: list[str]) -> None:
        self.base_pythons = base_pythons

    def __str__(self) -> str:
        return f"could not find python interpreter matching any of the specs {', '.join(self.base_pythons)}"



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Literal, cast

from packaging.utils import canonicalize_name

from tox.tox_env.errors import Fail

if TYPE_CHECKING:
    from collections.abc import Sequence

    from packaging._parser import Op, Value, Variable
    from packaging.markers import Marker, MarkerAtom, MarkerList
    from packaging.requirements import Requirement


def dependencies_with_extras(
    deps: list[Requirement],
    extras: set[str],
    package_name: str,
    *,
    available_extras: set[str] | None = None,
) -> list[Requirement]:
    return dependencies_with_extras_from_markers(
        extract_extra_markers(deps), extras, package_name, available_extras=available_extras
    )


def dependencies_with_extras_from_markers(
    deps_with_markers: list[tuple[Requirement, set[str | None]]],
    extras: set[str],
    package_name: str,
    *,
    available_extras: set[str] | None = None,
) -> list[Requirement]:
    normalized_extras = {canonicalize_name(e) for e in extras}
    if available_extras is not None and normalized_extras:
        normalized_available = {canonicalize_name(e) for e in available_extras}
        if unknown := normalized_extras - normalized_available:
            available_str = ", ".join(sorted(normalized_available)) or "none"
            unknown_str = ", ".join(sorted(unknown))
            msg = f"extras not found for package {package_name}: {unknown_str} (available: {available_str})"
            raise Fail(msg)
    result: list[Requirement] = []
    found: set[str] = set()
    todo: set[str | None] = normalized_extras | {None}
    visited: set[str | None] = set()
    while todo:
        new_extras: set[str | None] = set()
        for req, extra_markers in deps_with_markers:
            if todo & extra_markers:
                if req.name == package_name:  # support for recursive extras
                    new_extras.update(canonicalize_name(e) for e in (req.extras or set()))
                else:
                    req_str = str(req)
                    if req_str not in found:
                        found.add(req_str)
                        result.append(req)
        visited.update(todo)
        todo = new_extras - visited
    return result


def extract_extra_markers(deps: list[Requirement]) -> list[tuple[Requirement, set[str | None]]]:
    """Extract extra markers from dependencies.

    :param deps: the dependencies

    :returns: a list of requirement, extras set

    """
    return [_extract_extra_markers(d) for d in deps]


def _extract_extra_markers(req: Requirement) -> tuple[Requirement, set[str | None]]:
    req = deepcopy(req)
    markers: MarkerList = getattr(req.marker, "_markers", []) or []
    new_markers: MarkerList = []
    extra_markers: set[str] = set()
    marker = markers.pop(0) if markers else None
    while marker:
        extra = _get_extra(marker)
        if extra is not None:
            extra_markers.add(canonicalize_name(extra))
            if new_markers and new_markers[-1] in {"and", "or"}:
                del new_markers[-1]
            marker = markers.pop(0) if markers else None
            if marker in {"and", "or"}:
                marker = markers.pop(0) if markers else None
        else:
            new_markers.append(marker)
            marker = markers.pop(0) if markers else None
    if new_markers:
        cast("Marker", req.marker)._markers = new_markers  # noqa: SLF001
    else:
        req.marker = None
    return req, cast("set[str | None]", extra_markers) or {None}


def _get_extra(
    _marker: MarkerList | tuple[Variable | Value, Op, Variable | Value] | Sequence[MarkerAtom] | Literal["and", "or"],
) -> str | None:
    if not isinstance(_marker, tuple) or len(_marker) != 3:  # noqa: PLR2004
        return None
    marker_tuple = cast("tuple[Variable | Value, Op, Variable | Value]", _marker)
    left, op, right = marker_tuple
    if hasattr(left, "value") and left.value == "extra" and hasattr(op, "value") and op.value == "==":
        return right.value if hasattr(right, "value") else None
    return None



"""Load from a pyproject.toml file, native format."""

from __future__ import annotations

import sys
from collections.abc import Iterator, Mapping
from itertools import product
from typing import TYPE_CHECKING, Any, Final, cast

from tox.config.loader.section import Section
from tox.config.loader.toml import TomlLoader
from tox.config.loader.toml._product import expand_factor_group
from tox.config.types import MissingRequiredConfigKeyError
from tox.report import HandledError

from .api import Source

if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
    import tomllib
else:  # pragma: no cover (py311+)
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from tox.config.loader.api import Loader, OverrideMap
    from tox.config.sets import CoreConfigSet


class TomlSection(Section):
    SEP: str = "."
    PREFIX: tuple[str, ...]
    ENV: Final[str] = "env"
    ENV_BASE: Final[str] = "env_base"
    RUN_ENV_BASE: Final[str] = "env_run_base"
    PKG_ENV_BASE: Final[str] = "env_pkg_base"

    @classmethod
    def test_env(cls, name: str) -> TomlSection:
        return cls(cls.env_prefix(), name)

    @classmethod
    def env_prefix(cls) -> str:
        return cls.SEP.join((*cls.PREFIX, cls.ENV))

    @classmethod
    def core_prefix(cls) -> str:
        return cls.SEP.join(cls.PREFIX)

    @classmethod
    def package_env_base(cls) -> str:
        return cls.SEP.join((*cls.PREFIX, cls.PKG_ENV_BASE))

    @classmethod
    def run_env_base(cls) -> str:
        return cls.SEP.join((*cls.PREFIX, cls.RUN_ENV_BASE))

    @classmethod
    def env_base_prefix(cls) -> str:
        return cls.SEP.join((*cls.PREFIX, cls.ENV_BASE))

    @classmethod
    def env_base(cls, name: str) -> TomlSection:
        return cls(cls.env_base_prefix(), name)

    @property
    def keys(self) -> Iterable[str]:
        # Build keys from prefix + name directly, preserving dots in names (e.g. env "py3.11").
        prefix, name = self._prefix, self._name
        if prefix is None and not name:
            return []
        parts: list[str] = prefix.split(self.SEP) if prefix else []
        if self.PREFIX and len(parts) >= len(self.PREFIX) and tuple(parts[: len(self.PREFIX)]) == self.PREFIX:
            parts = parts[len(self.PREFIX) :]  # strip global PREFIX (e.g. ("tool", "tox"))
        if name:
            parts.append(name)
        return parts


class TomlPyProjectSection(TomlSection):
    PREFIX = ("tool", "tox")


class TomlPyProject(Source):
    """Configuration sourced from a pyproject.toml files."""

    FILENAME = "pyproject.toml"
    _Section: type[TomlSection] = TomlPyProjectSection

    def __init__(self, path: Path) -> None:
        if path.name != self.FILENAME or not path.exists():
            raise ValueError
        with path.open("rb") as file_handler:
            self._content = tomllib.load(file_handler)
        try:
            our_content: Mapping[str, Any] = self._content
            for key in self._Section.PREFIX:
                our_content = our_content[key]
            self._our_content = our_content
        except KeyError as exc:
            raise MissingRequiredConfigKeyError(path) from exc
        if set(self._our_content.keys()) == {"legacy_tox_ini"}:
            raise MissingRequiredConfigKeyError(path)
        self._env_base_generated: dict[str, str] = _build_env_base_map(
            dict(self._our_content.get(self._Section.ENV_BASE, {})),
        )
        super().__init__(path)

    def get_core_section(self) -> Section:
        return self._Section(prefix=None, name="")

    def transform_section(self, section: Section) -> Section:
        return self._Section(section.prefix, section.name)

    def get_loader(self, section: Section, override_map: OverrideMap) -> Loader[Any] | None:
        current = self._our_content
        sec = cast("TomlSection", section)
        for key in sec.keys:
            if key in current:
                current = current[key]
            else:
                return None
        if not isinstance(current, Mapping):
            msg = f"{sec.key} must be a table, is {current.__class__.__name__!r}"
            raise HandledError(msg)
        is_core = section.prefix is None
        is_env_base = not is_core and sec.prefix == self._Section.env_base_prefix()
        unused_exclude: set[str] = set()
        if is_core:
            unused_exclude = {sec.ENV, sec.ENV_BASE, sec.RUN_ENV_BASE, sec.PKG_ENV_BASE}
        elif is_env_base:
            unused_exclude = {"factors"}
        return TomlLoader(
            section=section,
            overrides=override_map.get(section.key, []),
            content=current,
            root_content=self._content,
            unused_exclude=unused_exclude,
        )

    def envs(self, core_conf: CoreConfigSet) -> Iterator[str]:
        yield from core_conf["env_list"]
        yield from [section.name for section in self.sections()]
        yield from self._env_base_generated

    def sections(self) -> Iterator[Section]:
        for env_name in self._our_content.get(self._Section.ENV, {}):
            if not isinstance(env_name, str):
                msg = f"Environment key must be string, got {env_name!r}"
                raise HandledError(msg)
            yield self._Section.test_env(env_name)

    def get_base_sections(self, base: list[str], in_section: Section) -> Iterator[Section]:
        core_prefix = self._Section.core_prefix()
        strip = f"{core_prefix}{self._Section.SEP}" if core_prefix else ""
        env_base_pfx = self._Section.env_base_prefix()
        env_base_dot = f"{env_base_pfx}{self._Section.SEP}"
        for entry in base:
            if entry.startswith(env_base_dot):
                yield self._Section.env_base(entry[len(env_base_dot) :])
            else:
                yield self._Section(prefix=core_prefix or None, name=entry.removeprefix(strip))
                if in_section.prefix is not None:
                    yield self._Section(prefix=in_section.prefix, name=entry)

    def get_tox_env_section(self, item: str) -> tuple[Section, list[str], list[str]]:
        if base_name := self._env_base_generated.get(item):
            return (
                self._Section.test_env(item),
                [self._Section.env_base_prefix() + self._Section.SEP + base_name, self._Section.run_env_base()],
                [self._Section.package_env_base()],
            )
        return self._Section.test_env(item), [self._Section.run_env_base()], [self._Section.package_env_base()]


def _build_env_base_map(env_base_content: dict[str, Any]) -> dict[str, str]:
    result: dict[str, str] = {}
    for base_name, config in env_base_content.items():
        if not isinstance(config, Mapping):
            msg = f"env_base.{base_name} must be a table"
            raise HandledError(msg)
        factors_raw = config.get("factors")
        if factors_raw is None:
            msg = f"env_base.{base_name} requires a 'factors' key; use [env.{base_name}] for single environments"
            raise HandledError(msg)
        if not isinstance(factors_raw, list):
            msg = f"env_base.{base_name}.factors must be a list, got {type(factors_raw).__name__}"
            raise HandledError(msg)
        if factors_raw and isinstance(factors_raw[0], list | dict):
            expanded = [expand_factor_group(g) for g in factors_raw]
            names = ["-".join(combo) for combo in product(*expanded)]
        else:
            names = [str(f) for f in factors_raw]
        for factor_suffix in names:
            result[f"{base_name}-{factor_suffix}"] = base_name
    return result


__all__ = [
    "TomlPyProject",
]



"""Defines the abstract base traits of a tox environment."""

from __future__ import annotations

import fnmatch
import logging
import os
import re
import string
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple, cast

from tox.execute.request import ExecuteRequest
from tox.tox_env.errors import Fail, Recreate, Skip
from tox.tox_env.info import Info
from tox.util.path import ensure_cachedir_tag, ensure_empty_dir, ensure_gitignore

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from io import BytesIO

    from tox.config.cli.parser import Parsed
    from tox.config.main import Config
    from tox.config.set_env import SetEnv
    from tox.config.sets import CoreConfigSet, EnvConfigSet
    from tox.execute.api import Execute, ExecuteStatus, Outcome, StdinSource
    from tox.journal import EnvJournal
    from tox.report import OutErr, ToxHandler
    from tox.tox_env.installer import Installer

LOGGER = logging.getLogger(__name__)
# Based on original gitleaks rule named generic-api-key
# See: https://github.com/gitleaks/gitleaks/blob/master/config/gitleaks.toml#L587
SECRET_KEYWORDS = [
    "access",
    "api",
    "auth",
    "client",
    "cred",
    "key",
    "passwd",
    "password",
    "private",
    "pwd",
    "secret",
    "token",
]
SECRET_ENV_VAR_REGEX = re.compile(
    r"""
    .*                  # any prefix
    ( {keywords} )      # one of the secret keywords
    .*                  # any suffix
    """.format(keywords="|".join(SECRET_KEYWORDS)),
    re.VERBOSE | re.IGNORECASE,
)


def redact_value(name: str, value: str) -> str:
    """Returns a redacted text if the key name looks like a secret."""
    if SECRET_ENV_VAR_REGEX.match(name):
        return "*" * len(value)
    return value


class ToxEnvCreateArgs(NamedTuple):
    """Arguments to pass on when creating a tox environment."""

    conf: EnvConfigSet
    core: CoreConfigSet
    options: Parsed
    journal: EnvJournal
    log_handler: ToxHandler


class ToxEnv(ABC):
    """A tox environment."""

    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        """Create a new tox environment.

        :param create_args: tox env create args

        """
        self.journal: EnvJournal = create_args.journal  #: handler to the tox reporting system
        self.conf: EnvConfigSet = create_args.conf  #: the config set to use for this environment
        self.core: CoreConfigSet = create_args.core  #: the core tox config set
        self.options: Parsed = create_args.options  #: CLI options
        self.log_handler: ToxHandler = create_args.log_handler  #: handler to the tox reporting system

        #: encode the run state of various methods (setup/clean/etc)
        self._run_state = {"setup": False, "clean": False, "teardown": False}
        self._paths_private: list[Path] = []  #: a property holding the PATH environment variables
        self._hidden_outcomes: list[Outcome] | None = []
        self._env_vars: dict[str, str] | None = None
        self._env_vars_pass_env: list[str] = []
        self._resolving_env_vars: bool = False
        self._suspended_out_err: OutErr | None = None
        self._execute_statuses: dict[int, ExecuteStatus] = {}
        self._interrupted = False
        self._log_id = 0

    @property
    def cache(self) -> Info:
        return Info(self.env_dir)

    @staticmethod
    @abstractmethod
    def id() -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def executor(self) -> Execute:
        raise NotImplementedError

    @property
    @abstractmethod
    def installer(self) -> Installer[Any]:
        raise NotImplementedError

    def _install(self, arguments: Any, section: str, of_type: str) -> None:
        from tox.plugin.manager import MANAGER  # noqa: PLC0415

        MANAGER.tox_on_install(self, arguments, section, of_type)
        self.installer.install(arguments, section, of_type)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.conf['env_name']})"

    def register_config(self) -> None:
        self.conf.add_constant(
            keys=["env_name", "envname"],
            desc="the name of the tox environment",
            value=self.conf.name,
        )
        self.conf.add_config(
            keys=["labels"],
            of_type=set[str],
            default=set(),
            desc="labels attached to the tox environment",
        )
        self.conf.add_config(
            keys=["env_dir", "envdir"],
            of_type=Path,
            default=lambda conf, name: cast("Path", conf.core["work_dir"]) / self.name,  # noqa: ARG005
            desc="directory assigned to the tox environment",
        )
        self.conf.add_config(
            keys=["env_tmp_dir", "envtmpdir"],
            of_type=Path,
            default=lambda conf, name: cast("Path", conf.core["work_dir"]) / self.name / "tmp",  # noqa: ARG005
            desc="a folder that is always reset at the start of the run",
        )
        self.conf.add_config(
            keys=["env_log_dir", "envlogdir"],
            of_type=Path,
            default=lambda conf, name: cast("Path", conf.core["work_dir"]) / self.name / "log",  # noqa: ARG005
            desc="a folder for logging where tox will put logs of tool invocation",
        )
        self.executor.register_conf(self)
        self.conf.default_set_env_loader = self._default_set_env
        self.conf.add_config(
            keys=["platform"],
            of_type=str,
            default="",
            desc="run on platforms that match this regular expression (empty means any platform)",
        )

        def pass_env_post_process(values: list[str]) -> list[str]:
            values.extend(self._default_pass_env())
            result = sorted(dict.fromkeys(values).keys())
            invalid_chars = set(string.whitespace)
            invalid = [v for v in result if any(c in invalid_chars for c in v)]
            if invalid:
                invalid_repr = ", ".join(repr(i) for i in invalid)
                msg = (
                    f"pass_env values cannot contain whitespace, use comma to have multiple values in a single line,"
                    f" invalid values found {invalid_repr}"
                )
                raise Fail(msg)
            return result

        self.conf.add_config(
            keys=["pass_env", "passenv"],
            of_type=list[str],
            default=[],
            desc="environment variables to pass on to the tox environment",
            post_process=pass_env_post_process,
        )
        self.conf.add_config(
            keys=["disallow_pass_env"],
            of_type=list[str],
            default=[],
            desc="environment variable patterns to exclude after pass_env glob expansion",
        )
        self.conf.add_config(
            "parallel_show_output",
            of_type=bool,
            default=False,
            desc="if set to True the content of the output will always be shown  when running in parallel mode",
        )
        self.conf.add_config(
            "recreate",
            of_type=bool,
            default=self._recreate_default,
            desc="always recreate virtual environment if this option is true, otherwise leave it up to tox",
        )
        self.conf.add_config(
            "allowlist_externals",
            of_type=list[str],
            default=[],
            desc="external command glob to allow calling",
        )
        assert self.installer is not None  # noqa: S101 # trigger installer creation to allow config registration

    def _recreate_default(self, conf: Config, value: str | None) -> bool:  # noqa: ARG002
        return cast("bool", self.options.recreate)

    @property
    def env_dir(self) -> Path:
        """:returns: the tox environments environment folder"""
        return cast("Path", self.conf["env_dir"])

    @property
    def env_tmp_dir(self) -> Path:
        """:returns: the tox environments temp folder"""
        return cast("Path", self.conf["env_tmp_dir"])

    @property
    def env_log_dir(self) -> Path:
        """:returns: the tox environments log folder"""
        return cast("Path", self.conf["env_log_dir"])

    @property
    def name(self) -> str:
        return cast("str", self.conf["env_name"])

    def _default_set_env(self) -> dict[str, str]:  # noqa: PLR6301
        return {}

    def _default_pass_env(self) -> list[str]:  # noqa: PLR6301
        env = [
            "https_proxy",  # HTTP proxy configuration
            "http_proxy",  # HTTP proxy configuration
            "no_proxy",  # HTTP proxy configuration
            "LANG",  # localization
            "LANGUAGE",  # localization
            "CURL_CA_BUNDLE",  # curl certificates
            "SSL_CERT_FILE",  # https certificates
            "CC",  # C compiler command
            "CFLAGS",  # C compiler flags
            "CCSHARED",  # compiler flags used to build a shared library
            "CXX",  # C++ compiler command
            "CPPFLAGS",  # C++ compiler flags
            "LD_LIBRARY_PATH",  # location of libs
            "LDFLAGS",  # linker flags
            "HOME",  # needed for `os.path.expanduser()` on non-Windows systems
            "FORCE_COLOR",  # force color output
            "NO_COLOR",  # disable color output
            "NETRC",  # used by pip and netrc modules
            "PYTHON_GIL",  # allows controlling python gil
        ]
        if sys.stdout.isatty():  # if we're on a interactive shell pass on the TERM
            env.append("TERM")
        if sys.platform == "win32":  # pragma: win32 cover
            env.extend(
                [
                    "TEMP",  # temporary file location
                    "TMP",  # temporary file location
                    "USERPROFILE",  # needed for `os.path.expanduser()`
                    "PATHEXT",  # needed for discovering executables
                    "MSYSTEM",  # controls paths printed format
                    "WINDIR",  # base path to system executables and DLLs
                ],
            )
        else:  # pragma: win32 no cover
            env.extend(
                [
                    "TMPDIR",  # temporary file location
                    "NIX_LD",  # nix-ld loader
                    "NIX_LD_LIBRARY_PATH",  # nix-ld library path
                    "SSH_AGENT_PID",  # ssh-agent process ID
                    "SSH_AUTH_SOCK",  # ssh-agent socket path
                ],
            )
        return env

    def setup(self) -> None:
        """Setup the tox environment."""
        if self._run_state["setup"] is False:  # pragma: no branch
            self._platform_check()
            recreate = cast("bool", self.conf["recreate"])
            if recreate:
                self._clean(transitive=True)
            try:
                self._setup_env()
                self._setup_with_env()
            except Recreate as exception:  # once we might try over
                if not recreate:  # pragma: no cover
                    logging.warning("recreate env because %s", exception.args[0])
                    self._clean(transitive=False)
                    self._setup_env()
                    self._setup_with_env()
            else:
                self._done_with_setup()
            finally:
                self._run_state["setup"] = True

    def teardown(self) -> None:
        if not self._run_state["teardown"]:
            try:
                self._teardown()
            finally:
                from tox.plugin.manager import MANAGER  # noqa: PLC0415

                MANAGER.tox_env_teardown(self)
                self._run_state["teardown"] = True

    def _teardown(self) -> None:  # noqa: B027 # empty abstract base class
        pass

    def _platform_check(self) -> None:
        """Skip env when platform does not match."""
        platform_str: str = self.conf["platform"]
        if platform_str:
            match = re.fullmatch(platform_str, self.runs_on_platform)
            if match is None:
                msg = f"platform {self.runs_on_platform} does not match {platform_str}"
                raise Skip(msg)

    @property
    @abstractmethod
    def runs_on_platform(self) -> str:
        raise NotImplementedError

    def _setup_env(self) -> None:
        """1. env dir exists
        2. contains a runner with the same type.

        """
        conf = {"name": self.conf.name, "type": type(self).__name__}
        with self.cache.compare(conf, ToxEnv.__name__) as (eq, old):
            if eq is False and old is not None:  # pragma: no branch  # recreate if already created and not equals
                msg = f"env type changed from {old} to {conf}"
                raise Recreate(msg)
        self._handle_env_tmp_dir()
        self._handle_core_tmp_dir()

    def _setup_with_env(self) -> None:  # noqa: B027 # empty abstract base class
        pass

    def _done_with_setup(self) -> None:  # noqa: B027 # empty abstract base class
        """Called when setup is done."""

    def _handle_env_tmp_dir(self) -> None:
        """Ensure exists and empty."""
        env_tmp_dir = self.env_tmp_dir
        if env_tmp_dir.exists() and next(env_tmp_dir.iterdir(), None) is not None:
            LOGGER.debug("clear env temp folder %s", env_tmp_dir)
            ensure_empty_dir(env_tmp_dir)
        env_tmp_dir.mkdir(parents=True, exist_ok=True)

    def _handle_core_tmp_dir(self) -> None:
        self.core["temp_dir"].mkdir(parents=True, exist_ok=True)
        ensure_cachedir_tag(self.core["work_dir"])
        ensure_gitignore(cast("Path", self.core["work_dir"]))

    def _clean(self, transitive: bool = False) -> None:  # noqa: ARG002, FBT001, FBT002
        if self._run_state["clean"]:  # pragma: no branch
            return  # pragma: no cover
        env_dir = self.env_dir
        if env_dir.exists():
            LOGGER.warning("remove tox env folder %s", env_dir)
            ensure_empty_dir(env_dir, except_filename="file.lock")
        self._log_id = 0  # we deleted logs, so start over counter
        self.cache.reset()
        self._run_state.update({"setup": False, "clean": True})

    @property
    def environment_variables(self) -> dict[str, str]:
        if self._resolving_env_vars:
            # Re-entrant call: set_env resolution triggered a substitution (e.g. {env_site_packages_dir})
            # that requires the virtualenv session, which needs environment_variables to be built first.
            # Return pass_env + PATH without set_env to break the cycle.
            result = self._load_pass_env(self.conf["pass_env"])
            result["PATH"] = self._make_path()
            return result

        pass_env: list[str] = self.conf["pass_env"]
        self._resolving_env_vars = True
        try:
            set_env: SetEnv = self.conf["set_env"]
        finally:
            self._resolving_env_vars = False
        if self._env_vars_pass_env == pass_env and not set_env.changed and self._env_vars is not None:
            return self._env_vars

        result = self._load_pass_env(pass_env)
        if disallow := self.conf["disallow_pass_env"]:
            disallow_patterns = [re.compile(fnmatch.translate(e), re.IGNORECASE) for e in disallow]
            result = {k: v for k, v in result.items() if not any(p.match(k) for p in disallow_patterns)}
        # load/paths_env might trigger a load of the environment variables, set result here, returns current state
        self._env_vars, self._env_vars_pass_env, set_env.changed = result, pass_env.copy(), False
        # set PATH here in case setting and environment variable requires access to the environment variable PATH
        result["PATH"] = self._make_path()
        self._resolving_env_vars = True
        try:
            for key in set_env:
                result[key] = set_env.load(key)
        finally:
            self._resolving_env_vars = False
        # if set_env modified PATH, re-prepend virtual-env paths (deduped) so they always come first
        if self._paths and "PATH" in set_env:
            result["PATH"] = self._make_path(result["PATH"])
        result["TOX_ENV_NAME"] = self.name
        result["TOX_WORK_DIR"] = str(self.core["work_dir"])
        result["TOX_ENV_DIR"] = str(self.conf["env_dir"])
        if (ci := os.environ.get("CI")) is not None:
            result["__TOX_ENVIRONMENT_VARIABLE_ORIGINAL_CI"] = ci
        return result

    @staticmethod
    def _load_pass_env(pass_env: list[str]) -> dict[str, str]:
        patterns = [re.compile(fnmatch.translate(e), re.IGNORECASE) for e in pass_env]
        result: dict[str, str] = {e: v for e, v in os.environ.items() if any(p.match(e) for p in patterns)}
        return result

    @property
    def _paths(self) -> list[Path]:
        return self._paths_private

    @_paths.setter
    def _paths(self, value: list[Path]) -> None:
        self._paths_private = value
        # Invalidate cached env vars so they rebuild on next access, preserving set_env PATH modifications.
        self._env_vars = None

    @property
    def _allow_externals(self) -> list[str]:
        result: list[str] = [f"{i}{os.sep}*" for i in self._paths]
        result.extend(i.strip() for i in self.conf["allowlist_externals"])
        return result

    def _make_path(self, existing: str | None = None) -> str:
        values = dict.fromkeys(str(i) for i in self._paths)
        values.update(dict.fromkeys((existing or os.environ.get("PATH", "")).split(os.pathsep)))
        return os.pathsep.join(values)

    def execute(  # noqa: PLR0913
        self,
        cmd: Sequence[Path | str],
        stdin: StdinSource,
        show: bool | None = None,  # noqa: FBT001
        cwd: Path | None = None,
        run_id: str = "",
        executor: Execute | None = None,
    ) -> Outcome:
        with self.execute_async(cmd, stdin, show, cwd, run_id, executor) as status:
            while status.wait() is None:
                pass  # pragma: no cover
        if status.outcome is None:  # pragma: no cover # this should not happen
            raise RuntimeError  # pragma: no cover
        return status.outcome

    def interrupt(self) -> None:
        """Interrupt the execution of a tox environment."""
        logging.warning("interrupt tox environment: %s", self.conf.name)
        self._interrupted = True
        for status in list(self._execute_statuses.values()):
            status.interrupt()

    @contextmanager
    def execute_async(  # noqa: PLR0913
        self,
        cmd: Sequence[Path | str],
        stdin: StdinSource,
        show: bool | None = None,  # noqa: FBT001
        cwd: Path | None = None,
        run_id: str = "",
        executor: Execute | None = None,
    ) -> Iterator[ExecuteStatus]:
        if self._interrupted:
            raise SystemExit(-2)  # pragma: no cover
        if cwd is None:
            cwd = self.core["tox_root"]
        if show is None:
            show = self.options.verbosity > 3  # noqa: PLR2004
        request = ExecuteRequest(cmd, cwd, self.environment_variables, stdin, run_id, allow=self._allow_externals)
        if request.cwd == _CWD:
            repr_cwd = ""
        else:
            try:
                repr_cwd = f" {_CWD.relative_to(cwd)}"
            except ValueError:
                repr_cwd = f" {cwd}"
        LOGGER.warning("%s%s> %s", run_id, repr_cwd, request.shell_cmd)
        out_err = self.log_handler.stdout, self.log_handler.stderr
        if executor is None:
            executor = self.executor
        with self._execute_call(executor, out_err, request, show) as execute_status:
            execute_id = id(execute_status)
            try:
                self._execute_statuses[execute_id] = execute_status
                yield execute_status
            finally:
                self._execute_statuses.pop(execute_id)
        if show and self._hidden_outcomes is not None and execute_status.outcome is not None:
            # if it gets canceled before even starting
            self._hidden_outcomes.append(execute_status.outcome)
        if self.journal and execute_status.outcome is not None:
            self.journal.add_execute(execute_status.outcome, run_id)
        self._log_execute(request, execute_status)

    def _log_execute(self, request: ExecuteRequest, status: ExecuteStatus) -> None:
        if self._log_id == 0:  # start with fresh slate on new run
            ensure_empty_dir(self.env_log_dir)
        self._log_id += 1
        self._write_execute_log(self.name, self.env_log_dir / f"{self._log_id}-{request.run_id}.log", request, status)

    @staticmethod
    def _write_execute_log(env_name: str, log_file: Path, request: ExecuteRequest, status: ExecuteStatus) -> None:
        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True, exist_ok=True)
        with log_file.open("wt", encoding="utf-8") as file:
            file.write(f"name: {env_name}\n")
            file.write(f"run_id: {request.run_id}\n")
            msg = ""
            for env_key, env_value in sorted(request.env.items()):
                redacted_value = redact_value(name=env_key, value=env_value)
                msg += f"env {env_key}: {redacted_value}\n"
            file.write(msg)
            for meta_key, meta_value in status.metadata.items():
                file.write(f"metadata {meta_key}: {meta_value}\n")
            file.write(f"cwd: {request.cwd}\n")
            allow = ["*"] if request.allow is None else request.allow
            file.write(f"allow: {':'.join(allow)}\n")
            file.write(f"cmd: {request.shell_cmd}\n")
            file.write(f"exit_code: {status.exit_code}\n")
        with log_file.open("ab") as file_b:
            if status.out:
                file_b.write(status.out)
            if status.err:  # pragma: no branch
                file_b.write(os.linesep.encode())
                file_b.write(b"standard error:")
                file_b.write(os.linesep.encode())
                file_b.write(status.err)

    @contextmanager
    def _execute_call(
        self,
        executor: Execute,
        out_err: OutErr,
        request: ExecuteRequest,
        show: bool,  # noqa: FBT001
    ) -> Iterator[ExecuteStatus]:
        with executor.call(
            request=request,
            env=self,
            show=show,
            out_err=out_err,
        ) as execute_status:
            yield execute_status

    @contextmanager
    def display_context(self, suspend: bool) -> Iterator[None]:  # noqa: FBT001
        with self._log_context(), self.log_handler.suspend_out_err(suspend, self._suspended_out_err) as out_err:
            if suspend:  # only set if suspended
                self._suspended_out_err = out_err
            yield

    def close_and_read_out_err(self) -> tuple[bytes, bytes] | None:
        if self._suspended_out_err is None:  # pragma: no branch
            return None  # pragma: no cover
        (out, err), self._suspended_out_err = self._suspended_out_err, None
        out_b, err_b = cast("BytesIO", out.buffer).getvalue(), cast("BytesIO", err.buffer).getvalue()
        out.close()
        err.close()
        return out_b, err_b

    @contextmanager
    def _log_context(self) -> Iterator[None]:
        with self.log_handler.with_context(self.conf.name):
            yield

    @property
    def _has_display_suspended(self) -> bool:
        return self._suspended_out_err is not None


_CWD = Path.cwd()



"""Pip requirements file parsing.

Specification is defined within pip itself and documented under:

- https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format
- https://github.com/pypa/pip/blob/master/src/pip/_internal/req/constructors.py#L291

"""

from __future__ import annotations



from __future__ import annotations

import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from packaging.pylock import Package, PylockValidationError
from packaging.pylock import Pylock as PackagingPylock
from packaging.requirements import Requirement

from tox.tox_env.errors import Fail

if TYPE_CHECKING:
    from pathlib import Path

if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib


@dataclass(frozen=True, kw_only=True)
class Pylock:
    path: Path
    extras: frozenset[str] = frozenset()
    groups: frozenset[str] = frozenset()
    marker_env: dict[str, str] = field(default_factory=dict)

    def requirements(self) -> list[Requirement]:
        with self.path.open("rb") as fh:
            data = tomllib.load(fh)
        try:
            parsed = PackagingPylock.from_dict(data)
        except PylockValidationError as exc:
            msg = f"invalid pylock file {self.path}: {exc}"
            raise Fail(msg) from exc
        env: dict[str, str | frozenset[str]] = {**self.marker_env}
        if self.extras:
            env["extras"] = self.extras
        if self.groups:
            env["dependency_groups"] = self.groups
        return [
            self._to_requirement(pkg)
            for pkg in parsed.packages
            if pkg.marker is None or pkg.marker.evaluate(env, context="lock_file")
        ]

    @staticmethod
    def _to_requirement(pkg: Package) -> Requirement:
        req_str = str(pkg.name)
        if pkg.version is not None:
            req_str += f"=={pkg.version}"
        return Requirement(req_str)


__all__ = [
    "Pylock",
]



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.config.sets import ConfigSet
    from tox.pytest import ToxProjectCreator
    from tox.session.state import State


def test_inline_tox_py(tox_project: ToxProjectCreator) -> None:
    def plugin() -> None:  # pragma: no cover # the code is copied to a python file
        import logging  # noqa: PLC0415

        from tox.plugin import impl  # noqa: PLC0415

        @impl
        def tox_add_option(parser: ToxParser) -> None:
            logging.warning("Add magic")
            parser.add_argument("--magic", action="store_true")

    project = tox_project({"toxfile.py": plugin})
    result = project.run("-h")
    result.assert_success()
    assert "--magic" in result.out


def test_toxfile_unknown_hook_skipped(tox_project: ToxProjectCreator) -> None:
    def plugin() -> None:  # pragma: no cover
        from tox.plugin import impl  # noqa: PLC0415

        @impl
        def tox_nonexistent_hook() -> None: ...

    project = tox_project({"toxfile.py": plugin, "tox.toml": ""})
    result = project.run("l")
    result.assert_success()
    assert "skipping inline plugin" in result.out


def test_toxfile_py_w_ephemeral_envs(tox_project: ToxProjectCreator) -> None:
    """Ensure additional ephemeral tox envs can be plugin-injected."""

    def plugin() -> None:  # pragma: no cover # the code is copied to a python file
        from tox.config.loader.memory import MemoryLoader  # noqa: PLC0415
        from tox.plugin import impl  # noqa: PLC0415

        env_name = "sentinel-env-name"

        @impl
        def tox_extend_envs() -> tuple[str]:
            return (env_name,)

        @impl
        def tox_add_core_config(core_conf: ConfigSet, state: State) -> None:  # noqa: ARG001
            in_memory_config_loader = MemoryLoader(
                base=["sentinel-base"],
                commands_pre=["sentinel-cmd"],
                description="sentinel-description",
            )
            state.conf.memory_seed_loaders[env_name].append(
                in_memory_config_loader,  # src/tox/provision.py:provision()
            )

    project = tox_project({"toxfile.py": plugin})

    tox_list_result = project.run("list", "-qq")
    tox_list_result.assert_success()
    expected_additional_env_txt = "\n\nadditional environments:\nsentinel-env-name -> sentinel-description"
    assert expected_additional_env_txt in tox_list_result.out

    tox_config_result = project.run("config", "-e", "sentinel-env-name", "-qq")
    tox_config_result.assert_success()
    assert "base = sentinel-base" in tox_config_result.out

    tox_run_result = project.run("run", "-e", "sentinel-env-name", "-q")
    tox_run_result.assert_failed()
    underlying_expected_oserror_msg = (
        "[WinError 2] The system cannot find the file specified"
        if sys.platform == "win32"
        else "[Errno 2] No such file or directory: 'sentinel-cmd'"
    )
    expected_cmd_lookup_error_txt = (
        f"sentinel-env-name: Exception running subprocess {underlying_expected_oserror_msg!s}\n"
    )
    assert expected_cmd_lookup_error_txt in tox_run_result.out



"""Print available tox environments."""

from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from tox.config.cli.parser import CORE
from tox.plugin import impl
from tox.session.env_select import register_env_select_flags

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("list", ["l"], "list environments", list_env, inherit=frozenset({CORE}))
    our.add_argument("--no-desc", action="store_true", help="do not show description", dest="list_no_description")
    d = register_env_select_flags(our, default=None, group_only=True)
    d.add_argument("-d", action="store_true", help="list just default envs", dest="list_default_only")


def list_env(state: State) -> int:
    option = state.conf.options
    has_group_select = bool(option.factors or option.labels)
    active_only = has_group_select or option.list_default_only

    active = dict.fromkeys(state.envs.iter())
    inactive = {} if active_only else {env: None for env in state.envs.iter(only_active=False) if env not in active}

    if not has_group_select and not option.list_no_description and active:
        print("default environments:")  # noqa: T201
    max_length = max((len(env) for env in chain(active, inactive)), default=0)

    def report_env(name: str) -> None:
        if not option.list_no_description:
            tox_env = state.envs[name]
            text = tox_env.conf["description"]
            if not text.strip():
                text = "[no description]"
            text = text.replace("\n", " ")
            msg = f"{env.ljust(max_length)} -> {text}".strip()
        else:
            msg = env
        print(msg)  # noqa: T201

    for env in active:
        report_env(env)

    if not has_group_select and not option.list_default_only and inactive:
        if not option.list_no_description:
            if active:  # pragma: no branch
                print()  # noqa: T201
            print("additional environments:")  # noqa: T201
        for env in inactive:
            report_env(env)
    return 0



from __future__ import annotations

from types import SimpleNamespace
from typing import TYPE_CHECKING

import pytest

from tox import __version__
from tox.plugin.manager import MANAGER

if TYPE_CHECKING:
    from pytest_mock import MockFixture

    from tox.pytest import ToxProjectCreator


def test_version() -> None:
    assert __version__


def test_version_without_plugin(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("--version")
    outcome.assert_success()
    assert __version__ in outcome.out
    assert "plugin" not in outcome.out


def test_version_with_plugin(tox_project: ToxProjectCreator, mocker: MockFixture) -> None:
    dist = [
        (
            mocker.create_autospec("types.ModuleType", __file__="B-path", tox_append_version_info=lambda: "magic"),
            SimpleNamespace(project_name="B", version="1.0"),
        ),
        (
            mocker.create_autospec("types.ModuleType", __file__="A-path"),
            SimpleNamespace(project_name="A", version="2.0"),
        ),
    ]
    mocker.patch.object(MANAGER.manager, "list_plugin_distinfo", return_value=dist)

    outcome = tox_project({"tox.ini": ""}).run("--version")

    outcome.assert_success()
    assert not outcome.err
    lines = outcome.out.splitlines()
    assert lines[0].startswith(__version__)

    assert lines[1:] == [
        "registered plugins:",
        "    B-1.0 at B-path magic",
        "    A-2.0 at A-path",
    ]


@pytest.mark.plugin_test
def test_version_with_inline_plugin(tox_project: ToxProjectCreator) -> None:
    def plugin() -> None:  # pragma: no cover
        pass

    project = tox_project({"tox.ini": "", "toxfile.py": plugin})
    outcome = project.run("--version")
    outcome.assert_success()
    assert __version__ in outcome.out
    assert "inline plugin:" in outcome.out
    assert "toxfile.py" in outcome.out


@pytest.mark.plugin_test
def test_version_with_inline_plugin_append_version_info(tox_project: ToxProjectCreator) -> None:
    def plugin() -> None:  # pragma: no cover
        def tox_append_version_info() -> str:
            return "custom-info-v1"

    project = tox_project({"tox.ini": "", "toxfile.py": plugin})
    outcome = project.run("--version")
    outcome.assert_success()
    assert "inline plugin:" in outcome.out
    assert "custom-info-v1" in outcome.out



from __future__ import annotations

import json
import shutil
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from tox.pytest import MonkeyPatch, ToxProjectCreator

ROOT = Path(__file__).parents[3]
SCHEMA_PATH = ROOT / "src" / "tox" / "tox.schema.json"


def test_show_schema_empty_dir(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)

    project = tox_project({})
    result = project.run("-qq", "schema")
    schema = json.loads(result.out)
    assert "properties" in schema
    assert "tox_root" in schema["properties"]


def test_schema_freshness(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(tmp_path)
    project = tox_project({
        "tox.toml": 'env_list = ["py"]',
        "pyproject.toml": '[build-system]\nrequires = ["setuptools"]\nbuild-backend = "setuptools.build_meta"',
    })
    result = project.run("-qq", "schema")
    generated = json.loads(result.out)
    committed = json.loads(SCHEMA_PATH.read_text())
    assert generated == committed, (
        "tox.schema.json is out of date — regenerate with: tox schema > src/tox/tox.schema.json"
    )


@pytest.mark.parametrize(
    ("filename", "tombi_cfg", "content"),
    [
        pytest.param(
            "tox.toml",
            '[[schemas]]\npath = "tox.schema.json"\ninclude = ["tox.toml"]\n',
            (ROOT / "tox.toml").read_text(),
            id="tox.toml",
        ),
        pytest.param(
            "pyproject.toml",
            '[[schemas]]\nroot = "tool.tox"\npath = "tox.schema.json"\ninclude = ["pyproject.toml"]\n',
            "[tool.tox]\nrequires = ['tox>=4']\nenv_list = ['py']\nskip_missing_interpreters = true\n\n"
            "[tool.tox.env_run_base]\ncommands = [['pytest']]\ndeps = 'pytest'\n\n"
            "[tool.tox.env.lint]\ncommands = [['ruff', 'check', '.']]\n",
            id="pyproject.toml",
        ),
    ],
)
def test_schema_tombi_lint(tmp_path: Path, filename: str, tombi_cfg: str, content: str) -> None:
    if not (tombi := shutil.which("tombi")):
        pytest.skip("tombi not installed")
    shutil.copy2(SCHEMA_PATH, tmp_path / "tox.schema.json")
    (tmp_path / filename).write_text(content)
    (tmp_path / "tombi.toml").write_text(tombi_cfg)
    result = subprocess.run(
        [tombi, "lint", "--error-on-warnings", "--offline", filename],
        cwd=tmp_path,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, f"{result.stdout}\n{result.stderr}"
